from django import template
from django.urls import reverse

register = template.Library()

from ..urls import absolute_reverse, absolutify


@register.simple_tag(takes_context=False, name="link.info")
def link_info(view_name, *targs, **kwargs):
    """Return complete link information dictionary for a Django view.

    This template tag returns a ShadowDict containing 'url', 'text', and
    'kwargs' for the specified view name. Use this when you need access to
    all link properties in your template logic.

    Args:
        view_name: Django URL pattern name to reverse
        *targs: Positional arguments - URL args followed by optional link text
        **kwargs: Additional HTML attributes for the link element

    Returns:
        ShadowDict: Dictionary with 'url', 'text', and 'kwargs' keys

    Example:
        {% load link %}
        {% link.info 'home' %}
        {% link.info 'article-detail' article.pk 'Read More' class='btn' %}
    """
    return link(view_name, *targs, **kwargs)


@register.simple_tag(takes_context=False, name="link.url")
def link_url(view_name, *targs, **kwargs):
    """Return only the URL string for a Django view.

    This template tag extracts just the 'url' value from the link dictionary,
    providing a shortcut when you only need the reversed URL path without
    additional link attributes.

    Args:
        view_name: Django URL pattern name to reverse
        *targs: Positional arguments passed to Django's reverse()
        **kwargs: Additional URL parameters (not used in URL generation)

    Returns:
        str: The reversed URL path

    Example:
        {% load link %}
        <a href="{% link.url 'home' %}">Home</a>
        <a href="{% link.url 'article-detail' article.pk %}">Article</a>
    """
    return link(view_name, *targs, **kwargs)["url"]


@register.inclusion_tag("trim/link.html", takes_context=True)
def abs_link(context, link, *targs, **kwargs):
    """Generate an absolute URL link with full domain.

    Converts a relative Django URL into an absolute URL including the protocol
    and domain from the current request context. Useful for emails, external
    references, or social media sharing links.

    Args:
        context: Django template context (automatically provided)
        link: Django URL pattern name to reverse
        *targs: Positional arguments - URL args followed by optional link text
        **kwargs: Additional HTML attributes for the link element

    Returns:
        dict: Link dictionary with absolute 'url', 'text', and 'kwargs'

    Example:
        {% load link %}
        {% abs_link 'article-detail' article.pk 'Read Article' %}
        {# Generates: https://example.com/articles/123/ #}
    """
    res = gen_link(link, *targs, **kwargs)
    res["url"] = absolutify(context["request"], res["url"])
    return res


@register.inclusion_tag("trim/link.html", takes_context=False)
def link(link, *targs, **kwargs):
    """Generate an HTML link element from a Django view name or existing link dict.

    This is the primary link template tag, supporting two modes:
    1. Create a new link from a view name (passes to gen_link)
    2. Update an existing ShadowDict link with new text or attributes

    The generated link is rendered using the 'trim/link.html' template.

    Args:
        link: Django URL pattern name (str) or existing ShadowDict to update
        *targs: Positional arguments - URL args followed by optional link text
        **kwargs: Additional HTML attributes (class, id, data-*, etc.)

    Returns:
        dict: Link dictionary with 'url', 'text', and 'kwargs' for template

    Examples:
        {% load link %}
        {# Basic link #}
        {% link 'home' 'Go Home' %}

        {# Link with URL arguments #}
        {% link 'article-detail' article.pk 'Read More' %}

        {# Link with HTML attributes #}
        {% link 'profile' user.id 'Profile' class='btn btn-primary' %}

        {# Update existing link #}
        {% link existing_link 'New Text' class='updated' %}
    """
    if isinstance(link, ShadowDict):
        if len(targs) > 0:
            link["text"] = targs[0]
        link["kwargs"].update(kwargs)
        return link
    return gen_link(link, *targs, **kwargs)


@register.inclusion_tag("trim/link.html", takes_context=False)
def new_link(link, *targs, **kwargs):
    """Generate a link that opens in a new tab with security attributes.

    Creates an HTML link with target='_blank' and rel='noreferrer noopener'
    by default. This is useful for external links or links you want to open
    in a new browser tab while maintaining security best practices.

    The 'noreferrer noopener' attributes prevent the new page from accessing
    the window.opener property and protect against tabnabbing attacks.

    Args:
        link: Django URL pattern name to reverse
        *targs: Positional arguments - URL args followed by optional link text
        **kwargs: Additional HTML attributes (can override target and rel)

    Returns:
        dict: Link dictionary with 'url', 'text', and 'kwargs' including
              target='_blank' and rel='noreferrer noopener'

    Examples:
        {% load link %}
        {# External link in new tab #}
        {% new_link 'external-resource' 'View Resource' %}

        {# Link with URL arguments #}
        {% new_link 'pdf-download' doc.id 'Download PDF' %}

        {# Override default attributes #}
        {% new_link 'popup' rel='noopener' class='external' %}
    """
    kwargs.setdefault("target", "_blank")
    kwargs.setdefault("rel", "noreferrer noopener")

    return gen_link(link, *targs, **kwargs)


@register.inclusion_tag("trim/js_link.html", name="js", takes_context=False)
def script_link(link, *targs, **kwargs):
    """Build a standard `<script>` for JS, usig the static path.

        {% link.js "examples/js/my-js-file.js"  %}

    Creating a standard script tag:

        <script type="text/javascript"  src="/static/examples/js/my-js-file.js></script>

    Apply args and kwargs as required:

        {% link.js "examples/js/petite-vue.iife.js" "defer" "init" %}
        # <script src="examples/js/petite-vue.iife.js" defer init></script>

    """
    kwargs.setdefault("type", "text/javascript")
    return {
        "static_name": link,
        "kwargs": kwargs,
        "args": targs,
    }


register.inclusion_tag("trim/js_link.html", name="link.js", takes_context=False)(
    script_link
)


@register.inclusion_tag("trim/css_link.html", name="css", takes_context=False)
def css_link(link, *targs, **kwargs):
    """

    Build a standard <link> for CSS, using a static path

        {% css 'css/main.css' %}

    All attributes of the node are available. As an example of versatility
    (not recommended):

        {% css 'images/favicon.ico' rel="shortcut icon" type="image/png" %}
    """
    kwargs.setdefault("rel", "stylesheet")
    kwargs.setdefault("type", "text/css")
    return {
        "static_name": link,
        "kwargs": kwargs,
        "args": targs,
    }


register.inclusion_tag("trim/css_link.html", name="link.css", takes_context=False)(
    css_link
)


@register.inclusion_tag("trim/link.html", takes_context=False)
def new_url_link(link, *targs, **kwargs):
    """Provide a _real_ URL, rather than a reversable name:

    {% load link %}
    {% for pay_url in flow.cart.receipt_links %}
        {% new_url_link pay_url "Receipt" %}
    {% endfor %}

    """
    kwargs.setdefault("target", "_blank")
    kwargs.setdefault("rel", "noreferrer noopener")
    kwargs.setdefault("url", link)
    return gen_link(link, *targs, **kwargs)


@register.inclusion_tag("trim/link.html", takes_context=False)
def url_link(link, *targs, **kwargs):
    """Provide a _real_ URL, rather than a reversable name:

    {% load link %}
    {% for pay_url in flow.cart.receipt_links %}
        {% url_link pay_url "Receipt" %}
    {% endfor %}

    """
    # kwargs.setdefault('target', '_blank')
    # kwargs.setdefault('rel', 'noreferrer noopener')
    kwargs.setdefault("url", link)
    return gen_link(link, *targs, **kwargs)


class ShadowDict(dict):
    pass


def gen_link(link, *targs, **kwargs):
    # "...X as name %}" detection
    hotix_as_word = targs[-2:] == (
        "",
        "",
    )
    ## We cut _3_ as (1) for the _last argument (text) and 2 for 'as X'
    ## else -1 for just the _last argument_
    cutlen = -3 if hotix_as_word else -1
    # Args are exerything given except the last "text" arg (and hotfix 'as X')
    args = targs[:cutlen] if len(targs) > 0 else targs
    url = kwargs.get("url", None) or reverse(link, args=args)
    text = targs[cutlen] if len(targs) > 0 else url

    return ShadowDict(
        url=url,
        text=text,
        kwargs=kwargs,
    )



from django import forms
from django.forms.widgets import CheckboxSelectMultiple
from wagtail.core.blocks import StreamBlock
from wagtailmarkdown.blocks import MarkdownBlock as WM_MarkdownBlock

from trim.wagtail import blocks as t_blocks


class Styles(t_blocks.StructBlock):

    styles = t_blocks.MultipleChoiceBlock(
        widget=CheckboxSelectMultiple,
        choices=(
            (
                "flipped",
                "Horizontal Flip",
            ),
            # ('b', 'b',),
        ),
    )

    def __init__(self, required=True, help_text=None, **kwargs):
        super().__init__(**kwargs)
        ss = self.declared_blocks["styles"]
        ch = kwargs.get("choices", (("egg", "Egg"),))
        ss.choices = ch

    class Meta:
        template = "blocks/styles.html"


from wagtail.core.blocks import StructValue


class LinkStructValue(StructValue):
    """
    https://docs.wagtail.org/en/v4.0.1/advanced_topics/customisation/streamfield_blocks.html#additional-methods-and-properties-on-structblock-values
    """

    def html(self):
        text = self.get("content")
        # page = self.get('page')
        return text or "No text for {self}"


class MarkdownBlock(t_blocks.StructBlock):
    content = StreamBlock(
        (
            (
                "text",
                WM_MarkdownBlock(icon="pilcrow"),
            ),
            (
                "document",
                t_blocks.document_chooser(icon="doc-full-inverse"),
            ),
        )
    )

    # text = WM_MarkdownBlock(icon="pilcrow")
    # document = t_blocks.document_chooser(icon="doc-full-inverse")
    css_styles = Styles()

    class Meta:
        template = "blocks/markdown-block.html"
        value_class = LinkStructValue



from collections import defaultdict

from django import template
from django.template import Context
from django.template.base import token_kwargs

from . import quickforms
from .shared_tools import parse_until

slot_space = defaultdict(dict)


def inject_node(parser):
    T = parser.tokens[0].__class__
    tr = "{{ slotplot }}"
    parser.prepend_token(T(parser.tokens[0].token_type, tr))


register = template.Library()


@register.tag(name="wrap")
def do_wrap(parser, token):
    """Wrap the contents given to the nodem with another template given through
    the tag token:

        {% wrap "stocks/wrap_form.html" with button_text="Next"  action="/fo/bar" %}
            <ul>
            {% for row in form %}
                <li>{{ row }}</li>
            {% endfor %}
            </ul>
        {% endwrap %}

    The template "wrap_forms.html" should refer to the `wrap.content` and any
    other variables from the `with` or parent contexts.

        <form method="post"
          enctype="multipart/form-data"
          {% if action %}action="{{action}}"{% endif %}>

            {{ wrap.content }}
            <input class='btn'
                type="submit"
                value="{% firstof button_text 'Confirm' %}">
            {% csrf_token %}
        </form>
    """
    nodelist, splits, extra = parse_until(parser, token, ("endwrap",))
    return WrappedContentNode(nodelist, splits, **extra)


class WrappedContentNode(template.Node):
    def __init__(self, nodelist, tokens, *a, **kw):
        self.nodelist = nodelist
        self.token_template_name = template.Variable(tokens[0])
        self.extra_context = kw

    def render(self, context):
        print(" rendering wrap", self)
        template_name = self.token_template_name.resolve(context)
        print("   Template name:", template_name)
        t = context.template.engine.get_template(template_name)
        ## The users given content.
        sub_ctx = {
            "wrap_key": id(self),
            "wrap_render_mode": "fragment",
        }
        with context.push(**sub_ctx):
            content = self.nodelist.render(context)

        # we extend the original context; allowing this unit to be overwitten
        # if required.
        wrap = {"content": content, "template_name": template_name}
        #  Dynamic key for the root context, static key for the 'wrap' object.
        content_key = "content"
        sub_ctx = {
            "wrap": wrap,
            content_key: content,
            # 'wrapper': context.setdefault('wrapper'),
            "wrap_render_mode": "render",
            "wrap_key": id(self),
        }

        values = {key: val.resolve(context) for key, val in self.extra_context.items()}
        with context.push(**sub_ctx, **values):
            print("    Rendering Wrapper content")
            # Context({'source': content}, autoescape=context.autoescape)
            return t.render(context)


@register.tag(name="slot")
def do_slot(parser, token):
    nodelist, splits, extra = parse_until(parser, token, ("endslot",))
    return SlotNode(nodelist, splits, **extra)


class SlotNode(template.Node):
    def __init__(self, nodelist, tokens, *a, **kw):
        self.nodelist = nodelist
        self.slot_name = template.Variable(tokens[0]) if len(tokens) > 0 else "noname"
        self.extra_context = kw

    def render(self, context):
        # template_name = self.token_template_name.resolve(context)
        # t = context.template.engine.get_template(template_name)
        wrm = context.get("wrap_render_mode", None)
        # wrapper = context.get('wrapper', None)
        kl = context.get("wrap_key")
        print(" rendering slot in mode:", wrm, kl)
        return getattr(self, f"render_mode_{wrm}")(context)

    def render_mode_render(self, context):
        content = self.nodelist.render(context)
        # we extend the original context; allowing this unit to e overbwitten
        # if required.
        wrap = {
            "content": content,
        }
        #  Dynamic key for the root context, static key for the 'wrap' object.
        content_key = "content"
        sub_ctx = {}  # = {'wrap':wrap, content_key:content}
        values = {key: val.resolve(context) for key, val in self.extra_context.items()}
        kl = context.get("wrap_key")
        with context.push(**sub_ctx, **values):
            #     # Context({'source': content}, autoescape=context.autoescape)
            #     return self.nodelist.render(context)
            nodelist_render = slot_space.get(kl, None)
            if nodelist_render:
                slotname = self.slot_name  # .resolve(context)
                return nodelist_render[slotname](context)
            else:
                # Apply slot default.
                return f"'no data for {kl}'"

    def render_mode_fragment(self, context):
        """In fragment mode the content is parsed from the wrap target. Capture
        the slots for later render by a wrapper.
        """
        content = self.nodelist.render(context)
        sub_ctx = {}

        kl = context.get("wrap_key")
        # In this case we don't render,instead capture under the wrapr name

        values = {key: val.resolve(context) for key, val in self.extra_context.items()}
        with context.push(**sub_ctx, **values):
            res = self.nodelist.render
            slot_space[context["wrap_key"]][self.slot_name or "no_slot_name"] = res
            # slot_space[context['wrap_key']]['default'] = content

        # slot_space[context['wrap_key']] = res
        return str(context["wrap_key"])


# @register.simple_tag(takes_context=True)
# def info_form(context, product_id):
#     return forms.ProductQuestionForm(initial={
#             'product_id':product_id,
#         })
#
# @register.inclusion_tag('dummy.html')
# def wrap(template='default.html'):
#     return {'template': template}



"""trim.models.fields.django

Django model fields as functions.
"""

from uuid import uuid4 as orig_uuid4

# from django.contrib.contenttypes.fields import GenericForeignKey
# from django.contrib.contenttypes.models import ContentType
from django.apps import apps
from django.contrib.auth import get_user_model as orig_get_user_model
from django.db import models

from trim import rand

sys_bool = bool
sys_int = int
sys_str = str

from .base import *


def url(*a, **kw):
    """A django `models.URLField` passing any arguments and keyword arguments

    provide as a standard field:

        class MyModel(models.Model):
            url = trims.url(max_length=2000)

    A CharField for a URL, validated by URLValidator.
    The default form widget for this field is a URLInput.
    Like all CharField subclasses, URLField takes the optional max_length
    argument. Ifyou don’t specify max_length, a default of 200 is used.

        class URLField(max_length=200, **options)
    """
    # kw.setdefault('max_length', 200)
    # defaults(kw, **blank_null())
    defaults(a, kw, nil=DEFAULT_NIL, max_length=200)
    return models.URLField(**kw)


def text(*a, **kw):
    """A standard `models.TextField` passing the standard arguments."""
    # defaults(kw, **blank_null())
    kw = defaults(a, kw, nil=DEFAULT_NIL)
    return models.TextField(*a, **kw)


def chars(first_var=None, *a, **kw):
    """A standard `models.CharField` passing the standard fields.
    The first value may be a callable `function` for the default value.
    """

    # If the first var is callable, it's the default func.
    # If the default= kwarg also exists, raise an error.
    default_max_length = 255
    max_length = default_max_length

    if callable(first_var):
        # Set as default
        max_length = kw.get("max_length", None)
        if kw.get("default", None) is not None:
            s = (
                "trims.chars(default_func, **params) accepts a"
                ' "default" keyword argument or a function as the first'
                " argument. Not both."
            )
            # Valid patterns:
            # chars(max_length=255)
            # chars(default_func)
            # chars(default=default_func)
            # chars(max_length=255, default=default_func)
            # Invalid:
            # chars(func, default=default_func)

            raise Exception(s)

        kw.setdefault("default", first_var)

    if max_length is None:
        max_length = default_max_length

    kw = defaults(a, kw, max_length=max_length, nil=DEFAULT_NIL)
    return models.CharField(*a, **kw)


def rand_str(*a, **kw):
    """Return a standard `CharField` with a default value of a random string
    using `.rand.rand_str`
    """
    kw = defaults(a, kw, default=rand.rand_str)
    return chars(*a, **kw)


def null_bool(*a, **kw):
    """A standard `BooleanField` with default `null=True`

    The default form widget for this field is NullBooleanSelect as null=True.
    The default value of BooleanField is None when Field.default isn’t defined.
    """
    kw = defaults(a, kw, nil=DEFAULT_NIL)
    return boolean(*a, **kw)


def true_bool(*a, **kw):
    """A standard `BooleanField` with default `True`

    The default form widget for this field is CheckboxInput, or NullBooleanSelect if null=True.
    The default value of BooleanField is None when Field.default isn’t defined.
    """
    kw = defaults(a, kw, default=True)
    return boolean(*a, **kw)


def false_bool(*a, **kw):
    """A standard `BooleanField` with default `False`

    The default form widget for this field is CheckboxInput, or NullBooleanSelect if null=True.
    The default value of BooleanField is None when Field.default isn’t defined.
    """
    kw = defaults(a, kw, default=False)
    return boolean(*a, **kw)


def boolean(*a, **kw):
    """A standard `BooleanField` passing all arguments.

    The default form widget for this field is CheckboxInput, or NullBooleanSelect if null=True.
    The default value of BooleanField is None when Field.default isn’t defined.
    """
    return models.BooleanField(*a, **kw)


def blank_dt(*a, **kw):
    """A standard `DateTimeField` with default `null=True`"""
    kw = defaults(a, kw, nil=DEFAULT_NIL)
    return datetime(*a, **kw)


def date(*a, **kw):
    """A standard `DateTimeField` passing all arguments"""
    kw = defaults(a, kw)
    return models.DateField(*a, **kw)


def dt_created(*a, **kw):
    """A standard `DateTimeField` with default `auto_now_add=True`"""
    kw.setdefault("auto_now_add", True)
    return datetime(*a, **kw)


def dt_updated(*a, **kw):
    """A standard `DateTimeField` with default `auto_now=True`"""
    kw.setdefault("auto_now", True)
    return datetime(*a, **kw)


def dt_cu_pair(*a, **kw):
    """A tuple pair of `DateTimeField`, first item `created`, second `updated`.
    `created, updated = dt_cu_pair()`

        class AnyModel(models.Model):
            name = chars()
            age = integer()
            created, updated = dt_cu_pair()

    """
    return (
        dt_created(*a, **kw),
        dt_updated(*a, **kw),
    )


def datetime(*a, **kw):
    """A standard `DateTimeField` passing all arguments"""
    kw = defaults(a, kw)
    return models.DateTimeField(*a, **kw)


def integer(*a, **kw):
    """A standard `IntegerField` with the first argument as the default
    `teenager_age = integer(18)`
    """
    value = None
    if len(a) > 0:
        value = a[0]
        a = a[1:]

    if value is not None:
        kw.setdefault("default", value)
    kw = defaults(a, kw)
    return models.IntegerField(*a, **kw)


def get_user_model():
    """Return the user model using the original django
    `django.contrib.auth.get_user_model` function.
    """
    return orig_get_user_model()


def fk(other, *a, on_delete=None, **kw):
    """A standard `ForeignKey` with the first argument as the _other_ model.
    `on_delete=CASCADE` by default.

        class EntityModel(models.Model):
            name = chars()
            attached = fk('other.model')
            owner = fk(get_user_model())
            color = fk(ColorModel)
    """
    kw = defaults(a, kw, on_delete=on_delete or models.CASCADE)
    return models.ForeignKey(other, *a, **kw)


def user_fk(*a, **kw):
    """A standard `ForeignKey` with the _other_ model as the `get_user_model()`
    result (the standard django auth user model).
    """
    kw = defaults(a, kw, nil=DEFAULT_NIL)
    return fk(get_user_model(), *a, **kw)


def self_fk(*a, **kw):
    """A standard `ForeignKey` with the _other_ model as the `"self"`"""
    kw = defaults(a, kw, nil=DEFAULT_NIL)
    return fk("self", *a, **kw)


def m2m(other, *a, **kw):
    """A standard `ManyToManyField` accepting the `other` reference of string
    to model type.
    """
    kw = defaults(a, kw, blank=True)
    if "null" in kw:
        # (fields.W340) null has no effect on ManyToManyField.
        kw.setdefault("blank", kw.pop("null"))
    return models.ManyToManyField(other, *a, **kw)


def self_m2m(*a, **kw):
    """A many to many field (through `m2m`) refering to _self_"""
    return m2m("self", *a, **kw)


def o2o(other, *a, on_delete=None, **kw):
    """A standard `OneToOneField` with the _other_ model as the first argument"""
    kw = defaults(a, kw, on_delete=on_delete or models.CASCADE)
    return models.OneToOneField(other, *a, **kw)


def user_o2o(*a, **kw):
    """A standard `OneToOneField` with the _other_ model as the `get_user_model()`
    result (the standard django auth user model).
    """
    return o2o(get_user_model(), *a, **kw)


def user_m2m(*a, **kw):
    """Create a many to many relationship to the user model,
    bound through get_user_model().
    """
    return m2m(get_user_model(), *a, **kw)


def image(*a, **kw):
    """A standard `ImageField` passing all arguments"""
    kw = defaults(a, kw, blank=True)
    return models.ImageField(*a, **kw)


def auto(*a, **kw):
    """A standard `AutoField` passing all arguments.

    An IntegerField that automatically increments according to available IDs.
    You usually won’t need to use this directly;
    a primary key field will automatically be added to your model if you
    don’t specify otherwise.

        class AutoField(**options)

    See Automatic primary key fields.
    """
    kw = defaults(a, kw)
    return models.AutoField(*a, **kw)


def big_auto(*a, **kw):
    """A 64-bit integer, much like an AutoField except that it is guaranteed to
    fit numbers from 1 to 9223372036854775807.

        class BigAutoField(**options)
    """
    kw = defaults(a, kw)
    return models.BigAutoField(*a, **kw)


def big_int(*a, **kw):
    """A 64-bit integer, much like an IntegerField except that it is guaranteed
    to fit numbers from -9223372036854775808 to 9223372036854775807.

        class BigIntegerField(**options)

    The default form widget for this field is a NumberInput.
    """
    kw = defaults(a, kw, nil=DEFAULT_NIL)
    return models.BigIntegerField(*a, **kw)


def binary(*a, bytes=2_097_152, **kw):
    """A standard `BinaryField` with the first argument as the target byte max"""
    # class BinaryField(max_length=None, **options)
    # A field to store raw binary data. It can be assigned bytes, bytearray, or memoryview.

    # By default, BinaryField sets editable to False, in which case it can’t be included in a ModelForm.

    # BinaryField has one extra optional argument:
    # BinaryField.max_length
    # The maximum length (in bytes) of the field.
    # The maximum length is enforced in Django’s validation using
    # MaxLengthValidator.
    kw = defaults((bytes,) + a, kw, nil=DEFAULT_NIL)
    return models.BinaryField(*a, **kw)


def decimal(*a, digits=19, places=10, **kw):
    """A standard `DecimalField` with defaults `max_digits=19` and
    `decimal_places=10`.


    A fixed-precision decimal number, represented in Python by a Decimal instance.
    It validates the input using DecimalValidator.

        class DecimalField(max_digits=None, decimal_places=None, **options)

    Has two required arguments:

    DecimalField.max_digits
        The maximum number of digits allowed in the number.
        Note that this number must be greater than or equal to decimal_places.

    DecimalField.decimal_places
        The number of decimal places to store with the number.

    For example, to store numbers up to 999.99 with a resolution of 2 decimal places,
    you’d use:
    models.DecimalField(..., max_digits=5, decimal_places=2)
    And to store numbers up to approximately one billion with a resolution of 10 decimal places:
    models.DecimalField(..., max_digits=19, decimal_places=10)
    """
    kw = defaults(a, kw, nil=DEFAULT_NIL)

    return models.DecimalField(*a, max_digits=digits, decimal_places=places, **kw)


def duration(*a, **kw):
    """A standard `DurationField`

    A field for storing periods of time - modeled in Python by timedelta.
    When used on PostgreSQL, the data type used is an interval and on
    Oracle the data type is INTERVAL DAY(9) TO SECOND(6).
    Otherwise a big_int of microseconds is used.

        class DurationField(**options)

    Note:

    Arithmetic with DurationField works in most cases. However on all
    databases other than PostgreSQL, comparing the value of a DurationField
    to arithmetic on DateTimeField instances will not work as expected.
    """

    kw = defaults(a, kw, nil=DEFAULT_NIL)
    return models.DurationField(*a, **kw)


def email(*a, **kw):
    """A CharField that checks that the value is a valid email address
    using EmailValidator.

        class EmailField(max_length=254, **options)
    """

    kw = defaults(a, kw, nil=DEFAULT_NIL)
    return models.EmailField(*a, **kw)


def file(*a, **kw):
    """A file-upload field.

    class FileField(upload_to=None, max_length=100, **options)
    """
    kw = defaults(a, kw, nil=DEFAULT_NIL)
    return models.FileField(*a, **kw)


def filepath(*a, **kw):
    """A CharField whose choices are limited to the filenames in a certain
    directory on the filesystem. Has some special arguments, of which the
    first is required:

        class FilePathField(path='', match=None, recursive=False,
            allow_files=True, allow_folders=False, max_length=100,
            **options)

    FilePathField.path
        Required.
        The absolute filesystem path to a directory from which this
        FilePathField should get its choices. Example: "/home/images".

    path may also be a callable, such as a function to dynamically set
    the path at runtime. Example:

        import os
        from django.conf import settings
        from django.db import models

        def images_path(*a, **kw):
            return os.path.join(settings.LOCAL_FILE_DIR, 'images')

        class MyModel(models.Model):
            file = models.FilePathField(path=images_path)
    """
    kw = defaults(a, kw, nil=DEFAULT_NIL)
    return models.FilePathField(*a, **kw)


def float_(*a, **kw):
    """A floating-point number represented in Python by a float instance.
    class FloatField(**options)

    The default form widget for this field is a NumberInput when localize is
    False or TextInput otherwise.

    """
    kw = defaults(a, kw, nil=DEFAULT_NIL)
    return models.FloatField(*a, **kw)


def ip_addr(*a, **kw):
    """
    An IPv4 or IPv6 address, in string format (e.g. 192.0.2.30 or 2a02:42fe::4).
    The default form widget for this field is a TextInput.

        class GenericIPAddressField(protocol='both', unpack_ipv4=False, **options)

    The IPv6 address normalization follows RFC 4291#section-2.2 section 2.2,
    including using the IPv4 format suggested in paragraph 3 of that section,
    like ::ffff:192.0.2.0. For example, 2001:0::0:01 would be normalized to
    2001::1, and ::ffff:0a0a:0a0a to ::ffff:10.10.10.10. All characters are
    converted to lowercase.

    GenericIPAddressField.protocol
        Limits valid inputs to the specified protocol. Accepted values are
        'both' (default), 'IPv4' or 'IPv6'. Matching is case insensitive.

    GenericIPAddressField.unpack_ipv4
        Unpacks IPv4 mapped addresses like ::ffff:192.0.2.1. If this option
        is enabled that address would be unpacked to 192.0.2.1. Default is
        disabled. Can only be used when protocol is set to 'both'.

    If you allow for blank values, you have to allow for null values since
    blank values are stored as null.
    """
    kw = defaults(a, kw, nil=DEFAULT_NIL)
    return models.GenericIPAddressField(*a, **kw)


def json(*a, **kw):
    """
    A field for storing JSON encoded data. In Python the data is represented
    in its Python native format: dictionaries, lists, strings, numbers,
    booleans and None.

        class JSONField(encoder=None, decoder=None, **options)

    JSONField is supported on MariaDB 10.2.7+, MySQL 5.7.8+, Oracle,
    PostgreSQL, and SQLite (with the JSON1 extension enabled).

    JSONField.encoder
        An optional json.JSONEncoder subclass to serialize data types not
        supported by the standard JSON serializer (e.g. datetime.datetime or
        UUID). For example, you can use the DjangoJSONEncoder class.

    Defaults to json.JSONEncoder.

    JSONField.decoder
        An optional json.JSONDecoder subclass to deserialize the value
        retrieved from the database. The value will be in the format chosen
        by the custom encoder (most often a string). Your deserialization
        may need to account for the fact that you can’t be certain of the
        input type. For example, you run the risk of returning a datetime
        that was actually a string that just happened to be in the same
        format chosen for datetimes.

    Defaults to json.JSONDecoder.

    If you give the field a default, ensure it’s an immutable object,
    such as a str, or a callable object that returns a fresh mutable
    object each time, such as dict or a function.
    Providing a mutable default object like default={} or default=[]
    shares the one object between all model instances.
    """
    kw = defaults(a, kw, nil=DEFAULT_NIL)
    return models.JSONField(*a, **kw)


def pos_big_int(*a, **kw):
    """Like a PositiveIntegerField, but only allows values under a certain
    (database-dependent) point. Values from 0 to 9223372036854775807 are safe
    in all databases supported by Django.

        class PositiveBigIntegerField(**options)
    """

    kw = defaults(a, kw, nil=DEFAULT_NIL)
    return models.PositiveBigIntegerField(*a, **kw)


def pos_int(*a, **kw):
    """Like an IntegerField, but must be either positive or zero (0). Values
    from 0 to 2147483647 are safe in all databases supported by Django. The
    value 0 is accepted for backward compatibility reasons.

        class PositiveIntegerField(**options)
    """

    kw = defaults(a, kw, nil=DEFAULT_NIL)
    return models.PositiveIntegerField(*a, **kw)


def pos_small_int(*a, **kw):
    """Like a PositiveIntegerField, but only allows values under a certain
    (database-dependent) point. Values from 0 to 32767 are safe in all
    databases supported by Django.

        class PositiveSmallIntegerField(**options)
    """
    kw = defaults(a, kw, max_length=50, nil=DEFAULT_NIL)
    return models.PositiveSmallIntegerField(*a, **kw)


def slug(*a, **kw):
    """Slug is a newspaper term. A slug is a trim label for something,
    containing only letters, numbers, underscores or hyphens. They’re generally
    used in URLs.

        class SlugField(max_length=50, **options)
    """
    kw = defaults(a, kw, max_length=50, nil=DEFAULT_NIL)
    return models.SlugField(*a, **kw)


def small_auto(*a, **kw):
    """Like an AutoField, but only allows values under a certain
    (database-dependent) limit. Values from 1 to 32767 are safe in all
    databases supported by Django.

        class SmallAutoField(**options)
    """

    kw = defaults(a, kw, nil=DEFAULT_NIL)
    return models.SmallAutoField(*a, **kw)


def small_int(*a, **kw):
    """Like an IntegerField, but only allows values under a certain
    (database-dependent) point. Values from -32768 to 32767 are safe in all
    databases supported by Django.

        class SmallIntegerField(**options)
    """

    kw = defaults(a, kw, nil=DEFAULT_NIL)
    return models.SmallIntegerField(*a, **kw)


def time(*a, **kw):
    """Return a standard `TimeField`

    A time, represented in Python by a datetime.time instance. Accepts the same auto-population options as DateField.

        class TimeField(auto_now=False, auto_now_add=False, **options)

    The default form widget for this field is a TimeInput. The admin adds some JavaScript trimcuts.
    """
    kw = defaults(a, kw, nil=DEFAULT_NIL)
    return models.TimeField(*a, **kw)


def uuid_null(*a, **kw):
    """Return a standard `UUIDField` through `uuid()`."""
    kw = defaults(a, kw, nil=DEFAULT_NIL)
    return uuid(*a, **kw)


def pk_uuid(*a, **kw):
    """Return a standard `UUIDField` field through `uuid()`, with the
    `primary_key=True` and the default parameter as the function `uuid.uuid4`

    Example:

        class ProfileModel(models.Model):
            pk = fields.pk_uuid()
            user = fields.user_fk()

    """
    # return models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    return uuid(primary_key=True, default=orig_uuid4, editable=False)


def uuid(*a, **kw):
    """Return a standard `UUIDField`, with editable=False and the default
    function as `uuid.uuid4()`.

    example:

        class MyUUIDModel(models.Model):
            my_uuid = trim.models.fields.uuid()

    A field for storing universally unique identifiers.
    Uses Python’s UUID class.

    When used on PostgreSQL, this stores in a uuid datatype, otherwise in
    a char(32).

        class UUIDField(**options)

    Universally unique identifiers are a good alternative to AutoField for
    primary_key. The database will not generate the UUID for you, so it is
    recommended to use default:

        import uuid
        from django.db import models

        class MyUUIDModel(models.Model):
            id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
            # other fields


    Note that a callable (with the parentheses omitted) is passed to default,
    not an instance of UUID.

    """
    kw = defaults(a, kw, editable=False, default=orig_uuid4)
    return models.UUIDField(*a, **kw)


def str_uuid(*a, **kw):
    """Return a standard `CharField` through `chars()`, the default value
    as the function `uuid.uuid4()`.
    """
    kw = defaults(a, kw, default=orig_uuid4)
    return chars(*a, **kw)


class LazyImport:
    def get_GenericForeignKey(self):
        from django.contrib.contenttypes.fields import GenericForeignKey

        return GenericForeignKey

    def get_ContentType(self):
        from django.contrib.contenttypes.models import ContentType

        # from django.contrib.contenttypes.fields import GenericForeignKey
        return ContentType

    def __getitem__(self, name):
        return getattr(self, f"get_{name}")()


GEN_C = {"lazy": LazyImport()}


def get_cached(name):
    v = GEN_C.get(name, None)
    if v is None:
        v = GEN_C[name] = GEN_C["lazy"][name]
    return v


def contenttype_fk(content_type=None, *a, **kw):

    # content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    is_str = isinstance(content_type, sys_str)
    ContentType = get_cached("ContentType")
    content_type = apps.get_model(content_type) if is_str else content_type
    content_type = ContentType if content_type is None else content_type
    kw = defaults(a, kw, on_delete=models.CASCADE)
    return fk(content_type, **kw)


CONTENT_TYPE_FIELD = "content_type"
ID_FIELD = "object_id"


def generic_fk(content_type_field=CONTENT_TYPE_FIELD, id_field=ID_FIELD, **kw):
    """Return the standard GenericForeignKey, providing a content type field and
    an id field `generic_fk('content_type', 'object_id')`

    Synonymous to:

        class ExampleModel(models.Model):
            # content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
            # object_id = models.PositiveIntegerField()
            content_object = GenericForeignKey('content_type', 'object_id')

    Example:

        class ExampleModel(models.Model):
            content_type = contenttype_fk()
            object_id = pos_int()
            content_object = generic_fk('content_type', 'object_id')

    """
    kw = defaults(None, kw)
    GenericForeignKey = get_cached("GenericForeignKey")

    return GenericForeignKey(content_type_field, id_field)


def add_generic_key(model, field, content_type_field=None, id_field=None, **kw):
    owner, content_type, object_id = any_model_set(**kw)
    setattr(model, field, owner)
    content_type_field = content_type_field or f"{field}_content_type"
    pk_field = id_field or f"{field}_object_id"
    setattr(model, content_type_field, content_type)
    setattr(model, pk_field, object_id)
    return model


def any(prefix, content_type_field=None, id_field=None, **kw):
    """Allow a shorter syntax for calling any_model_set by assigning a 'prefix'
    to the default content_type and object_id associated fields names.

    before:

       (unit,
        unit_content_type,
        unit_object_id) = fields.any_model_set('unit_content_type',
                                               'unit_object_id'
                                               )

    after:

        unit, unit_content_type, unit_object_id = fields.any(prefix='unit')

    """
    prefix = prefix or ""
    pr_u = f"{prefix}_" if len(prefix) > 0 else ""
    content_type_field = content_type_field or f"{pr_u}{CONTENT_TYPE_FIELD}"
    id_field = id_field or f"{pr_u}{ID_FIELD}"

    return any_model_set(content_type_field, id_field, **kw)


def any_model_set(*a, nil=True, **kw):
    """Return a tuple of three fields, `generic_fk`, `contenttype_fk`, and `pos_int`
    for a set of _content type_ associations.

    Apply a generic foreignkey insertion on a model using three fields:

        GenericForeignKey   the field to manipulate: e.g. "Product.parent_entity"
        FK: ContentType     The FK content_type, to store the ContentType model filter
        Pos Int             The key of the target model.

    Apply to the class as three arguments,

        class Product(models.Model):
            owner, content_type, object_id = trims.any_model_set()

    The name of `content_type` and `object_id` should match the key values
    through the GenericForeignKey:


        class Product(models.Model):
            ( hobby,
              hobby_content_type,
              hobby_pk) = trims.any_model_set(
                                             "hobby_content_type",
                                             "hobby_pk"
                                             )

    The first two parameters for the `any_model_set` function should match the
    fields collected within the class:

        class Product(models.Model):
            owner, owner_content_type, owner_object_id = trims.any_model_set(
                                    # owner Generif FK target attributes.
                                    'owner_content_type', 'owner_object_id')

            (hobby,
             hobby_content_type,
             hobby_object_id) = trims.any_model_set(
                                        'hobby_content_type',
                                        'hobby_object_id'
                                    )


    Use the field through the primary field name, in this case `owner`:

        item = Product.objects.get(100)
        item.owner = Other.objects.get(id=200)
        item.save()
    """
    return (generic_fk(*a, nil=nil), contenttype_fk(nil=nil), pos_int(nil=nil))


float = float_

auto_small = small_auto
auto_big = big_auto
int_small = small_int
int_big = big_int
int_pos = pos_int
int_small_pos = pos_small_int
int_big_pos = pos_big_int
int_ = integer
int = integer

str = string = chars

boolean_null = bool_null = null_bool
boolean_true = bool_true = true_bool
boolean_false = bool_false = false_bool

bool = boolean

dt_blank = blank_dt
dt = datetime

any_model = any

o2o_user = user_o2o
fk_user = user_fk
fk_self = self_fk
m2m_self = self_m2m

created = dt_created
updated = dt_updated



from . import errors
from .auth import *
from .base import *
from .download import streamfile_response
from .errors import Custom404
from .list import OrderPaginatedListView
from .serialized import JsonDetailView, JsonListView, JSONResponseMixin, JsonView



from collections import defaultdict

classes = defaultdict(tuple)


def hook_model_mixin_class(cls):
    """Given an autoloaded AutoModelMixin, stash the model against the target
    parent model. These mixings are applied at loadout.
    """
    # print('Register', cls)
    model_name = cls.Meta.model_name

    if not isinstance(model_name, str):
        _m = model_name._meta
        model_name = f"{_m.model.__module__}{_m.object_name}"
    classes[model_name] += (cls,)


def get_classes():
    return classes


class AutoModelMixin(object):
    """Apply to a class, as an extension of the target class:


    from trim.models import AutoModelMixin

    class HyperlinkList(AutoModelMixin):

        def get_hyperlinks(self):
            return Hyperlink.get_user_models(self)

        class Meta:
            # The target app 'baskets' and its model 'Cart'
            model_name = fields.get_user_model() #  'otherapp.Person'
    """

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        hook_model_mixin_class(cls)


def hook_init_model_mixins(sender):
    # print('Hook init', sender)
    meta = sender._meta
    an = meta.app_label
    mn = meta.model_name
    on = meta.object_name

    short_model = f"{an}.{on}"
    module_model = f"{meta.model.__name__}.{meta.model.__module__}"
    model_name = f"{meta.model.__module__}{meta.object_name}"
    g = lambda x: classes.get(x, None) or ()
    lists = g(short_model) + g(module_model) + g(on) + g(model_name)
    # print(" -- > hooks: ", short_model, module_model, on, model_name)
    # print(" -- > ", len(lists))
    bind_mixins(sender, lists)


def bind_mixins(sender, lists):
    for _class in lists:
        for k in dir(_class):
            if k.startswith("_"):
                continue
            v = getattr(_class, k)
            setattr(sender, k, v)


def hook_waiting_model_mixins():
    # print('Hook models')
    # Get the classes,
    # append methods to models
    pass



"""Tools to assist with dynamic loadout of settings within trim and the
client site.
"""


class LiveConfigure(object):

    def installed_apps(self):
        """Return a list of strings to add to the INSTALLED_APPS"""
        return []

    def middleware(self):
        return []

    def configure_settings(self, name):
        print("configure_settings", name)



import inspect
import sys
from pathlib import Path

from django.apps import apps
from django.contrib.auth import get_user_model
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.shortcuts import render
from django.urls import reverse, reverse_lazy
from django.views.generic import (
    CreateView,
    DeleteView,
    DetailView,
    FormView,
    ListView,
    TemplateView,
    UpdateView,
    View,
)
from django.views.generic.dates import (
    ArchiveIndexView,
    DateDetailView,
    DayArchiveView,
    MonthArchiveView,
    TodayArchiveView,
    WeekArchiveView,
    YearArchiveView,
)

Permissioned = PermissionRequiredMixin


from trim import names as trim_names

ALL = "__ALL__"


class ShortMixin:

    def get_template_names(self):
        v = super().get_template_names()
        mapped_name = trim_names.get_mapped_name(self)
        p = Path(v[0])
        name = p.stem.replace(f"{self.model.__name__.lower()}_", "")
        v += [
            # user general overide.
            f"crud/{name}.html",
            f"crud/{mapped_name}.html",
            # Builtins
            f"trim/crud/{name}.html",
            f"trim/crud/{mapped_name}.html",
        ]
        return v

    def get_context_data(self, **kwargs):

        kwargs.setdefault("model_name", self.model.__name__)
        return super().get_context_data(**kwargs)


def extract_location(target):
    # appname == 'products'
    # mod_first == 'hyperlink'
    # name == 'Hyperlink'
    name = target.__name__
    mod_first = first_bit(name).lower()
    appname = first_bit(target.__module__)
    return (
        appname,
        mod_first,
        name,
    )


def create_class_slot(master_class, *additional_classes, **kwargs):
    _cls = (
        ShortMixin,
        master_class,
    ) + additional_classes
    return (
        _cls,
        (
            kwargs,
            {},
        ),
    )


def first_bit(word):
    return word.split(".")[0]


def discover_models(target_name, models=None, module_needles=None):
    """
    trims.guess_classes(models=grab_models(models))
    trims.guess_classes()
    """

    if models is not None:
        return models

    no_needles = module_needles in (
        False,
        ALL,
    )
    needles = () if module_needles is None else module_needles
    keeps = ()

    model_name = first_bit(target_name)
    all_classes = apps.get_models(model_name)
    needles += (model_name,)

    for model in all_classes:
        module_name = first_bit(model.__module__)
        if no_needles or (module_name in needles):
            keeps += (model,)

    return keeps


def crud_classes(
    target_name=None,
    model_class=None,
    success_url=None,
    success_url_bit="list",
    models=None,
    module_needles=None,
    **base_definition,
):
    """
    return tuple(
            crud(m, target_name, success_url, success_url_bit, **base_definition)
            for m in ensure_tuple(model_class)
        )
    """
    r = ()
    if target_name is None:
        target_name = inspect.currentframe().f_back.f_globals["__name__"]

    if model_class is None:
        model_class = discover_models(target_name, models, module_needles)

    model_class = ensure_tuple(model_class)

    for m in model_class:
        v = crud(
            m,
            class_module_name=target_name,
            success_url=success_url,
            success_url_bit=success_url_bit,
            **base_definition,
        )
        r += (v,)
    return r


def history_classes(
    target_name=None,
    model_class=None,
    models=None,
    class_module_name=None,
    module_needles=None,
    **base_definition,
) -> tuple:
    """Generate a range of "history" class-based-views for the given `model_class`
    list. This is synonymous to calling `trim.views.history` repeatedly.

    Given a `model_class` as a `models.Model`, `list` or `tuple` type, build a
    set of Archive based views into the `target_name`.

    Arguments:
        **base_definition {dict} -- attributtes given to all classes as base
                                    class properties and methods

    Keyword Arguments:
        target_name {str} -- Name of the target module to insert the newly generated
            classes. If `None` the `__name__` of the calling method module
            is applied. (default: {None})
        model_class {models.Model, tuple, list} -- The target model or models
            to generate views. If `None` dicover the models using
            `trim.views.discover_models` (default: {None})
        models {list, tuple} -- Use an explicit list of `models` over `model_class`.
            If `None` the model_class is used (default: {None})
        class_module_name {str} -- The string name of the module to insert the
            newly generated classes, such as `"products.views"`.
            If `None` attempt to capture the _last frame_ caller module name,
            defaulting to the calling module name. (default: {None})
        module_needles {list, tuple} -- A list of module names to focus upon
            if model discovery is used. If a dicovered model originated from
            a module name within the needles, history views will be created
            (default: {None}).

    Returns:
        {tuple} -- A tuple of generated classes. The class-based-views already
                 exist within the `target_name`
    """

    if target_name is None:
        target_name = inspect.currentframe().f_back.f_globals["__name__"]

    if model_class is None:
        model_class = discover_models(target_name, models, module_needles)

    model_class = ensure_tuple(model_class)

    r = ()
    for m in model_class:
        v = history(
            m, class_module_name=class_module_name or target_name, **base_definition
        )
        r += (v,)
    return r


def crud(
    model,
    class_module_name=None,
    success_url=None,
    success_url_bit="list",
    **base_definition,
):
    """
    crud(models.Product)


    class ProductListView(ListView):
        model = models.Product

    class ProductCreateView(CreateView):
        model = models.Product
        fields = '__all__'


    class ProductUpdateView(UpdateView):
        model = models.Product
        fields = '__all__'


    class ProductDeleteView(DeleteView):
        model = models.Product
        success_url = reverse_lazy('products:list')


    class ProductDetailView(DetailView):
        model = models.Product


    same as:

        listview = gen_class(name, (ListView,), base_definition)
        createview = gen_class(name, (CreateView,), base_definition, **create_update_def)
        updateview = gen_class(name, (UpdateView,), base_definition, **create_update_def)
        detailview = gen_class(name, (DetailView,), base_definition)
        deleteview = gen_class(name, (DeleteView,), base_definition, **success_url_d)

    returns:

        return (
            listview,
            detailview,
            createview,
            updateview,
            deleteview,
        )
    """
    appname, mod_first, name = extract_location(model)

    lazy_url = success_url or f"{appname}:{mod_first}-{success_url_bit}"
    base_definition.setdefault("model", model)
    success_url_d = {"success_url": reverse_lazy(lazy_url)}
    create_update_def = {"fields": "__all__", **success_url_d}

    parts = (
        # Order is important here.
        # ( (ShortMixin, CreateView, ), (base_definition, create_update_def), ),
        ListView,
        (
            CreateView,
            create_update_def,
        ),
        (
            UpdateView,
            create_update_def,
        ),
        DetailView,
        (DeleteView, success_url_d),
    )

    return thin_parts_gen(parts, name, base_definition, class_module_name)


def history(model, class_module_name=None, **base_definition):
    """
    trims.history(models.Product, __name__, date_field='created')
    """

    if class_module_name is None:
        class_module_name = inspect.currentframe().f_back.f_globals["__name__"]

    appname, mod_first, name = extract_location(model)
    base_definition.setdefault("model", model)
    base_definition.setdefault("date_field", "created")

    parts = (
        ArchiveIndexView,
        DateDetailView,
        DayArchiveView,
        MonthArchiveView,
        TodayArchiveView,
        WeekArchiveView,
        YearArchiveView,
    )

    return thin_parts_gen(parts, name, base_definition, class_module_name)


def thin_parts_gen(parts, name, base_definition, class_module_name):
    packs = gen_thin_packs(parts, base_definition)
    return gen_packed_views(name, class_module_name, packs)


def gen_thin_packs(parts, base_definition):
    """
    parts = (
        # Order is important here.
        # ( (ShortMixin, CreateView, ), (base_definition, create_update_def), ),
        (ListView,),
        (CreateView, create_update_def,),
        (UpdateView, create_update_def,),
        (DetailView,),
        (DeleteView, success_url_d),
    )


    packs = (
        # Order is important here.
        # ( (ShortMixin, CreateView, ), (base_definition, create_update_def), ),
        create_class_slot(ListView, **base_definition),
        create_class_slot(CreateView, **base_definition, **create_update_def),
        create_class_slot(UpdateView, **base_definition, **create_update_def),
        create_class_slot(DetailView, **base_definition),
        create_class_slot(DeleteView, **base_definition, **success_url_d),
    )
    """
    packs = ()
    for part in parts:
        part = ensure_tuple(part)
        kw = part[1] if len(part) > 1 else {}
        slot = create_class_slot(part[0], **base_definition, **kw)
        packs += (slot,)
    return packs


def gen_packed_views(name, class_module_name, view_packs, master_class_position=1):
    """
    packs = (
        # Order is important here.
        # ( (ShortMixin, CreateView, ), (base_definition, create_update_def), ),
        create_class_slot(ListView, **base_definition),
        create_class_slot(CreateView, **base_definition, **create_update_def),
        create_class_slot(UpdateView, **base_definition, **create_update_def),
        create_class_slot(DetailView, **base_definition),
        create_class_slot(DeleteView, **base_definition, **success_url_d),
    )
    """

    r = ()
    for parents, definitions in view_packs:

        viewclass = gen_class(
            name,
            parents,
            definitions[0],
            class_module_name=class_module_name,
            master_class_position=master_class_position,
            **definitions[1],
        )
        r += (viewclass,)

    return r


def ensure_tuple(items):
    if isinstance(items, (list, tuple)) is False:
        items = (items,)
    return items


def gen_class(
    crud_name,
    crud_parents,
    class_definition,
    class_module_name=None,
    master_class_position=-1,
    **params,
):

    if class_module_name is None:
        class_module_name = __name__

    crud_parents = ensure_tuple(crud_parents)
    safe = params.pop("safe", True)

    parent_class_name = crud_parents[master_class_position].__name__
    view_class_name = (
        params.pop("view_class_name", None) or f"{crud_name}{parent_class_name}"
    )
    module = sys.modules[class_module_name]

    view_cache[view_class_name] += 1

    if hasattr(module, view_class_name) is True and (safe is True):
        print("EXISTS", view_class_name)
        return getattr(module, view_class_name)

    class_members = copy_update(class_definition, **params)
    new_view_class = type(view_class_name, crud_parents, class_members)

    # print('Generating', view_class_name)
    setattr(module, view_class_name, new_view_class)

    return new_view_class


from collections import defaultdict

view_cache = defaultdict(int)


def copy_update(entity, **params):
    r = entity.copy()
    r.update(params)
    return r



# from wagtail.admin.edit_handlers import StreamFieldPanel
from wagtail.admin.edit_handlers import FieldPanel, InlinePanel, MultiFieldPanel
from wagtail.images.edit_handlers import ImageChooserPanel

# from wagtail.snippets.edit_handlers import SnippetChooserPanel

StreamFieldPanel = FieldPanel
SnippetChooserPanel = FieldPanel
ImageChooserPanel = FieldPanel


def image(*a, **kw):
    return ImageChooserPanel(*a, **kw)


def streamfield(*a, **kw):
    return StreamFieldPanel(*a, **kw)


def field(*a, **kw):
    return FieldPanel(*a, **kw)


def inline(*a, **kw):
    return InlinePanel(*a, **kw)


def snippet(*a, **kw):
    return SnippetChooserPanel(*a, **kw)



"""
Test trim.merge module.

Simple tests for file merging utility.
"""

import unittest
from pathlib import Path
from tempfile import TemporaryDirectory

from trim.merge import FileExists, recombine, split_i


class SplitITest(unittest.TestCase):
    """Test split_i helper function."""

    def test_extracts_part_number(self):
        # Setup & Execute
        result = split_i("filename.part_0")

        # Assert
        self.assertEqual(result, 0)

    def test_extracts_higher_part_number(self):
        # Setup & Execute
        result = split_i("filename.part_42")

        # Assert
        self.assertEqual(result, 42)

    def test_sorts_correctly(self):
        # Setup
        files = ["file.part_2", "file.part_0", "file.part_1"]

        # Execute
        sorted_files = sorted(files, key=split_i)

        # Assert
        self.assertEqual(sorted_files, ["file.part_0", "file.part_1", "file.part_2"])


class FileExistsTest(unittest.TestCase):
    """Test FileExists exception."""

    def test_can_raise_exception(self):
        # Setup & Execute & Assert
        with self.assertRaises(FileExists):
            raise FileExists("/some/path")

    def test_stores_message(self):
        # Setup
        path = "/test/path"

        # Execute
        exc = FileExists(path)

        # Assert
        self.assertEqual(exc.args[0], path)


class RecombineTest(unittest.TestCase):
    """Test recombine function."""

    def test_merges_split_files(self):
        # Setup
        with TemporaryDirectory() as tmpdir:
            tmppath = Path(tmpdir)

            # Create test split files
            (tmppath / "testfile.part_0").write_bytes(b"Hello ")
            (tmppath / "testfile.part_1").write_bytes(b"World")

            # Execute
            output = recombine(tmppath)

            # Assert
            self.assertTrue(output.exists())
            self.assertEqual(output.read_bytes(), b"Hello World")

    def test_raises_if_output_exists(self):
        # Setup
        with TemporaryDirectory() as tmpdir:
            tmppath = Path(tmpdir)

            # Create split files and output file
            (tmppath / "testfile.part_0").write_bytes(b"data")
            (tmppath / "testfile").write_bytes(b"existing")

            # Execute & Assert
            with self.assertRaises(FileExists):
                recombine(tmppath)


if __name__ == "__main__":
    unittest.main()



# Generated by Django 4.0.2 on 2022-10-19 00:59

import uuid

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("account", "0006_alter_accountemail_user"),
    ]

    operations = [
        migrations.AddField(
            model_name="accountemail",
            name="uuid_token",
            field=models.CharField(
                blank=True, default=uuid.uuid4, max_length=255, null=True
            ),
        ),
    ]



# Generated by Django 4.0.2 on 2022-05-15 17:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("account", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="AccountEmail",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(blank=True, max_length=255, null=True)),
                (
                    "email_address",
                    models.EmailField(blank=True, max_length=254, null=True),
                ),
                ("verified", models.DateTimeField(blank=True, null=True)),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.AddField(
            model_name="account",
            name="verified",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="account",
            name="associated_emails",
            field=models.ManyToManyField(blank=True, to="account.AccountEmail"),
        ),
    ]



from django.template.base import token_kwargs


def parse_until(parser, token, ends):
    """Gather all information within the context of the parser until the `ends`
    are met:


        @register.tag(name="markdown")
        def do_slot(parser, token):
            nodelist, splits, extra = parse_until(parser, token, ('endmarkdown',))
            return MarkdownContentNode(nodelist, splits, **extra)

    """
    nodelist = parser.parse(ends)
    parser.delete_first_token()
    # rather than delete, change;
    # inject_node(parser)
    splits = token.split_contents()[1:]
    extra = extract_with_statement(token, parser, splits=splits)
    return nodelist, splits, extra


def extract_with_statement(token, parser, splits=None):
    """
    Discover anything after the _with_ statement, and return a dict.
    """

    extra = {}
    splits = splits or token.split_contents()[1:]
    for i, word in enumerate(splits):
        if word == "with":
            extra = token_kwargs(splits[i + 1 :], parser)
    return extra


from django.template.loader_tags import TemplateSyntaxError


def parse_tag(parser, token):
    """REturn bits, the _with_ context,"""

    bits = token.split_contents()

    res_bits = []
    options = {}
    tag_name, *remaining_bits = bits
    while remaining_bits:
        option = remaining_bits.pop(0)
        if option in options:
            raise TemplateSyntaxError(
                "The %r option was specified more than once." % option
            )
        value = None
        if option == "with":
            value = token_kwargs(remaining_bits, parser, support_legacy=False)
            if not value:
                raise TemplateSyntaxError(
                    '"with" in %r tag needs at least one keyword argument.' % bits[0]
                )
            options[option] = value
        else:
            # raise TemplateSyntaxError(
            #     "Unknown argument for %r tag: %r." % (bits[0], option)
            # )
            res_bits.append(option)

    namemap = options.get("with", {})

    return res_bits, namemap, options



import argparse
import shlex
from functools import partial
from pathlib import Path

from trim import execute

from .base import AppActions, AppArgument, AppFunction, ConfigMixin
from .base import Help as BHelp
from .base import SubHelpFormatter, print_help

# def print_help(parser):

#     # retrieve subparsers from parser
#     subparsers_actions = [
#         action for action in parser._actions
#         if isinstance(action, argparse._SubParsersAction)]
#     # there will probably only be one subparser_action,
#     # but better safe than sorry
#     for subparsers_action in subparsers_actions:
#         # get all subparsers and print help

#         for choice, subparser in subparsers_action.choices.items():
#             print("--- Subparser '{}'".format(choice))
#             subparser.formatter_class = SubHelpFormatter
#             print(subparser.format_help())


class Help(BHelp):
    double_dash = (
        "apply a double-switch before the " "input to ensure switch aggregation."
    )
    pattern_append = "Append a partial string to the filename split pattern"


class GraphApps(object):
    verbose = 0

    def build_graph_parsers(self, top, parser=None, depth=0, position=None):
        """Given a nested tree of parsers build sub apps into the argparser

        {
            "graph": {
                "keys": [
                    "download",
                ],
                "download": {
                    "depth": 0,
                    "keys": ["db"],
                    "position": ["db"],
                    "items": [
                        "download"
                    ],
                    "db": {
                        "depth": 1,
                        "keys": [],
                        "position": [],
                        "items": [
                            "db"
                        ]
                    }
                }
            },
            "db": "py scripts/download_remote.py"
        }

        trim download db
        """
        for key in top.get("keys", ()):
            data = top[key]
            inner_keys = data.get("keys", ())
            position = tuple(data.get("position", ()))
            position = (position or ()) + (key,)
            if self.verbose > 0:
                w = " " * depth
                print(w, "Adding", position)
            # build a caller
            func = partial(self.depthed_default_caller, data)
            # Push into the parsers
            sub_parser = self.add_sub(
                func, key, parser, formatter_class=SubHelpFormatter
            )
            # recurse.
            self.build_graph_parsers(
                data, sub_parser, depth=depth + 1, position=position
            )

    def depthed_default_caller(self, content, args=None):
        items = content.get("items", ())
        conf = self.get_conf()
        datas = {x: conf.get(x) for x in items}
        self.namespace = args
        self.run_units(datas)
        return "res from default_caller"

    def run_units(self, datas):
        # print('\nKey', datas)
        # print('  default_caller: Key', self.namespace)
        for i, (k, d) in enumerate(datas.items()):
            self.execute_step(i, k, d)

    def execute_step(self, i, k, d):
        pass


class StepExecute(object):

    def execute_step(self, i, k, d):
        # Shell true, because the user is building the scripts.
        command = shlex.split(d)
        command += self.namespace.unknown_args
        print("  Running", k, " -> ", command)
        execute.read_one_stream_command(command)


class TrimApp(AppActions, StepExecute, GraphApps):
    """TrimApp extends AppActions to present provide trim app functions
    for the CLI

    """

    prog_name = "trim"
    primary_name = "trim"

    def setup(self):
        conf = self.get_conf()
        graph = conf.get("graph", {})
        self.verbose = conf.get("verbose", 0)
        self.build_graph_parsers(graph)

    def scripts_func(self, args):
        print("scripts_func", args)
        return "res from scripts_func"

    def default_caller(self, args):
        print_help(self.get_primary_parser())
        return "res from default_caller"

    def add(self, *app_function):
        self.app_functions += app_function


class TrimAdminApp(TrimApp):
    register_name = "admin"

    def get_conf(self):
        return {}


class VerboseSwitch(AppArgument):

    def setup_args(self, parser):
        parser.add_argument("--verbose", "-v", action="count", default=0)


class VersionSwitch(AppArgument):
    _version = "1.0"
    target = "primary"

    def setup_args(self, parser):
        parser.add_argument(
            "--version", action="version", version=f"%(prog)s {self._version}"
        )


class Scripts(AppFunction):
    # name = 'scripts'
    help = Help.scripts
    register_name = "admin"


class ScriptsAddFilenameArg(AppArgument):
    """The 'filename' argument for the `scripts add` function.

    This is applied through the `arguments` of ScriptsAdd.

    As this defines early, it will fail the primary hook parser.
    """

    target = "scripts.add"

    def setup_args(self, parser):
        parser.add_argument(
            "filename", type=str, nargs="*", help=Help.add_filename, default="."
        )
        parser.add_argument(
            "--", dest="_", nargs="+", type=str, default=None, help=Help.double_dash
        )
        parser.add_argument(
            "--type",
            type=str,
            default="raw",
            choices=(
                "raw",
                "path",
            ),
        )
        parser.add_argument(
            "-n",
            "--name",
            type=str,
            default=None,
        )
        parser.add_argument(
            "-d",
            "--dry",
            action="store_true",
        )
        parser.add_argument(
            "-f",
            "--dry-full",
            action="store_true",
        )
        parser.add_argument("-p", "--pattern", type=str, default=None)
        parser.add_argument(
            "-pa", "--pattern-append", type=str, default=None, help=Help.pattern_append
        )


import re
from pprint import pprint as pp


class DefaultHelp(AppArgument):
    target = "primary"

    def setup_args(self, parser):
        # parser.add_help = False
        del parser._option_string_actions["--help"]
        del parser._option_string_actions["-h"]
        parser.add_argument("-h", "--help", action="help")
        # parser.set_defaults(func=self.caller)

    def caller(self, *args, **kwargs):
        print("HELP", args, kwargs)

    # def prep(self, parser):
    #     print('DefaultHelp.prep')
    #     parser.primary_init_kwargs['add_help'] = False
    #     return super().prep(parser)


class ScriptInstall(ConfigMixin):
    """Install a script given through the CLI by file."""

    def parse_script_path_place(self, parsed, **kwargs):
        return parsed.filename.absolute().as_posix()

    def parse_raw_path_place(self, parsed, **kwargs):
        return str(parsed.filename)

    def as_path(self, parts):
        return Path("".join(parts))

    def install(self, parsed, **kwargs):
        path = getattr(self, f"parse_{parsed.type}_path_place")(parsed, **kwargs)

        name = parsed.name or self.as_path(parsed.filename).stem
        parts = self.parse_name(name, parsed.pattern_append, parsed.pattern)
        print("Installing", name)
        data = self.get_conf()
        if parsed.dry:
            r = self.append_graph(data if parsed.dry_full else {}, parts, name)
            r[name] = path
            pp(r)
            return
        self.append_graph(data, parts, name)
        data[name] = path
        self.write_conf_data(data)

    def append_graph(self, data, parts, name, graph=None):

        graph = graph or data.get("graph", None)
        if "graph" not in data:
            data["graph"] = {"keys": []}
        leaf = data["graph"]

        for i, part in enumerate(parts):
            if part not in leaf:
                print("Add leaf", part)
                leaf[part] = {"depth": i, "keys": [], "position": parts[0:i]}
            keys = set(leaf["keys"])
            keys.add(part)
            leaf["keys"] = tuple(keys)
            leaf = leaf[part]

        items = set(leaf.get("items") or [])
        items.add(name)
        leaf["items"] = tuple(items)
        return data

    def parse_name(self, name, append_pattern="", pattern=None):
        safe_ap = append_pattern or ""
        pattern = pattern or f"[_ -{safe_ap}]"
        return re.split(pattern, name)


class ScriptsAdd(AppFunction):
    """The "scripts add [filename]" command for installing a command into
    the persistent store.
    """

    parent_name = "scripts"
    name = "add"
    help = Help.add
    register_name = "admin"

    arguments = (ScriptsAddFilenameArg,)

    def hook(self, parsed, *args, **kwargs):
        print("hook", parsed)
        si = ScriptInstall()
        si.install(parsed, **kwargs)
        return "res from hook"


actions = TrimApp()
# actions.add(VerboseSwitch, VersionSwitch, Scripts, ScriptsAdd, ScriptsAddFilenameArg)
actions.prep()

admin_actions = TrimAdminApp()
admin_actions.add(AppActions, Scripts)
# admin_actions.add(VerboseSwitch, VersionSwitch, Scripts, ScriptsAdd, ScriptsAddFilenameArg)
admin_actions.prep()


def main():
    print("Entry from main")
    actions.run_hook()


def main_admin():
    print("Admin entry from main")
    admin_actions.run_hook()



from django.forms import widgets
from django.forms.widgets import *


def checkboxes(*a, **kw):
    """A standard `widgets.CheckboxSelectMultiple` widget."""
    return widgets.CheckboxSelectMultiple(*a, **kw)


def textarea(*a, **kw):
    """A standard `widgets.Textarea` widget."""
    return widgets.Textarea(*a, **kw)


def password(*a, **kw):
    """A standard `widgets.PasswordInput` widget."""
    return widgets.PasswordInput(*a, **kw)


def radios(*a, **kw):
    """A standard `widgets.RadioSelect` widget."""
    return widgets.RadioSelect(*a, **kw)


def hidden(*a, **kw):
    """A standard `widgets.HiddenInput` widget."""
    return widgets.HiddenInput(*a, **kw)


def checkbox(*a, **kw):
    """CheckboxInput"""
    return widgets.CheckboxInput(*a, **kw)


checkbox_select_multiple = checkboxes


def choice(*a, **kw):
    """ChoiceWidget"""
    return widgets.ChoiceWidget(*a, **kw)


class MultipleFileInput(widgets.ClearableFileInput):
    allow_multiple_selected = True


def clearable_file_input(*a, **kw):
    """ClearableFileInput"""
    return MultipleFileInput(*a, **kw)


def date(*a, **kw):
    """DateInput"""
    return widgets.DateInput(*a, **kw)


def date_time_base(*a, **kw):
    """DateTimeBaseInput"""
    return widgets.DateTimeBaseInput(*a, **kw)


def date_time(*a, **kw):
    """DateTimeInput"""
    return widgets.DateTimeInput(*a, **kw)


def email(*a, **kw):
    """EmailInput"""
    return widgets.EmailInput(*a, **kw)


def file(*a, **kw):
    """FileInput"""
    return widgets.FileInput(*a, **kw)


def base(*a, **kw):
    """Input"""
    return widgets.Input(*a, **kw)


def multi_widget(*a, **kw):
    """MultiWidget"""
    return widgets.MultiWidget(*a, **kw)


def multiple_hidden(*a, **kw):
    """MultipleHiddenInput"""
    return widgets.MultipleHiddenInput(*a, **kw)


def null_boolean_select(*a, **kw):
    """NullBooleanSelect"""
    return widgets.NullBooleanSelect(*a, **kw)


def number(*a, **kw):
    """NumberInput"""
    return widgets.NumberInput(*a, **kw)


def ordered_set(*a, **kw):
    """OrderedSet"""
    return widgets.OrderedSet(*a, **kw)


def select(*a, **kw):
    """Select"""
    return widgets.Select(*a, **kw)


def select_date(*a, **kw):
    """SelectDateWidget"""
    return widgets.SelectDateWidget(*a, **kw)


def select_multiple(*a, **kw):
    """SelectMultiple"""
    return widgets.SelectMultiple(*a, **kw)


def split_date_time(*a, **kw):
    """SplitDateTimeWidget"""
    return widgets.SplitDateTimeWidget(*a, **kw)


def split_hidden_date_time(*a, **kw):
    """SplitHiddenDateTimeWidget"""
    return widgets.SplitHiddenDateTimeWidget(*a, **kw)


def text(*a, **kw):
    """TextInput"""
    return widgets.TextInput(*a, **kw)


input = text
chars = text


def textarea(*a, **kw):
    """Textarea"""
    return widgets.Textarea(*a, **kw)


area = textarea


def time(*a, **kw):
    """TimeInput"""
    return widgets.TimeInput(*a, **kw)


def url(*a, **kw):
    """URLInput"""
    return widgets.URLInput(*a, **kw)



from django.shortcuts import render

from .custom404 import *


def handler404(request, exception):
    # page_not_found_view
    return render(request, "trim/errors/404.html", {"status_code": 404})


def handler500(request, exception=None):
    # error_view
    return render(request, "trim/errors/500.html", {"status_code": 500})


def handler403(request, exception=None):
    # permission_denied_view
    return render(request, "trim/errors/403.html", {"status_code": 403})


def handler400(request, exception=None):
    # bad_request_view
    return render(request, "trim/errors/400.html", {"status_code": 400})



# https://blog.ovalerio.net/archives/1512
from django import template

register = template.Library()


@register.simple_tag(takes_context=True)
def updated_params(context, **kwargs):
    """Update URL query parameters while preserving existing ones.

    Copies all GET parameters from the current request and updates them
    with the provided keyword arguments, returning a URL-encoded query string.
    This is particularly useful for pagination, filtering, and sorting links
    where you want to maintain existing query parameters.

    Args:
        context: Django template context (automatically provided)
        **kwargs: Query parameters to add or update

    Returns:
        str: URL-encoded query string ready to use in href attributes

    Examples:
        {% load updated_params %}

        {# Basic pagination - preserves existing filters #}
        <a href="?{% updated_params page=2 %}">Next Page</a>

        {# Update multiple parameters #}
        <a href="?{% updated_params page=3 sort='name' order='asc' %}">Page 3</a>

        {# Change filter while keeping pagination #}
        <a href="?{% updated_params category='electronics' %}">Electronics</a>

    Reference:
        Credit to O. Valerio: https://blog.ovalerio.net/archives/1512
    """
    res = context["request"].GET.copy()
    res.update(kwargs)
    return res.urlencode()



from pathlib import Path

from django.http import FileResponse


def content_type_response(filepath, ext=None, content_types=None, default=None):
    """Return a `FileResponse` type with the Content-Disposition applied.

    from trim.response import content_type_response

    real_filepath = Path('real/file.js')
    return content_type_response(real_filepath)

    """
    default = default or "application/octet-stream"

    content_type_map = content_types or {
        "png": "image/png",
        "pdf": "application/pdf",
        "unknown": default,
    }

    content_type = content_type_map.get(Path(filepath).suffix[1:], default)

    response = FileResponse(filepath.open("rb"), content_type=content_type)

    response["Content-Disposition"] = "filename={}".format(filepath.name)
    return response


def content_data_response(filedata, filename, content_types=None, default=None):
    """Return a `FileResponse` type with the Content-Disposition applied.

    from trim.response import content_type_response

    real_filedata = Path('real/file.js')
    return content_type_response(real_filedata)

    """
    default = default or "application/octet-stream"

    content_type_map = content_types or {
        "png": "image/png",
        "pdf": "application/pdf",
        "unknown": default,
    }

    content_type = content_type_map.get(Path(filename).suffix[1:], default)

    response = FileResponse(filedata, content_type=content_type)
    response["Content-Disposition"] = "filename={}".format(filename)
    return response



import shlex
import subprocess
import sys


def read_one_stream_command(command):
    """Open a cli command, executing the task waiting for a returncode completion
    Inputs are processed through the pip.

      subprocess.STD_INPUT_HANDLE
          The standard input device. Initially, this is the console input
          buffer, CONIN$.
      subprocess.STD_OUTPUT_HANDLE
          The standard output device. Initially, this is the active console
          screen buffer, CONOUT$.
      subprocess.STD_ERROR_HANDLE
          The standard error device. Initially, this is the active console
          screen buffer, CONOUT$.
      subprocess.SW_HIDE
          Hides the window. Another window will be activated.
      subprocess.STARTF_USESTDHANDLES
          Specifies that the STARTUPINFO.hStdInput, STARTUPINFO.hStdOutput,
          and STARTUPINFO.hStdError attributes contain additional information.
      subprocess.STARTF_USESHOWWINDOW
          Specifies that the STARTUPINFO.wShowWindow attribute contains
          additional information.
      subprocess.CREATE_NEW_CONSOLE
          The new process has a new console, instead of inheriting its
          parent’s console (the default).
      subprocess.CREATE_NEW_PROCESS_GROUP
          A Popen creationflags parameter to specify that a new process
          group will be created. This flag is necessary for using os.kill()
          on the subprocess.
          This flag is ignored if CREATE_NEW_CONSOLE is specified.
      subprocess.ABOVE_NORMAL_PRIORITY_CLASS
          A Popen creationflags parameter to specify that a new process will
          have an above average priority.
      subprocess.BELOW_NORMAL_PRIORITY_CLASS
          A Popen creationflags parameter to specify that a new process will
          have a below average priority.
      subprocess.HIGH_PRIORITY_CLASS
          A Popen creationflags parameter to specify that a new process will
          have a high priority.
      subprocess.IDLE_PRIORITY_CLASS
          A Popen creationflags parameter to specify that a new process will
          have an idle (lowest) priority.
      subprocess.NORMAL_PRIORITY_CLASS
          A Popen creationflags parameter to specify that a new process will
          have an normal priority. (default)
      subprocess.REALTIME_PRIORITY_CLASS
          A Popen creationflags parameter to specify that a new process will
          have realtime priority. You should almost never use
          REALTIME_PRIORITY_CLASS, because this interrupts system threads
          that manage mouse input, keyboard input, and background disk
          flushing. This class can be appropriate for applications that
          “talk” directly to hardware or that perform brief tasks that
          should have limited interruptions.
      subprocess.CREATE_NO_WINDOW
          A Popen creationflags parameter to specify that a new process will
          not create a window.
      subprocess.DETACHED_PROCESS
          A Popen creationflags parameter to specify that a new process will
          not inherit its parent’s console. This value cannot be
          used with CREATE_NEW_CONSOLE.
      subprocess.CREATE_DEFAULT_ERROR_MODE
          A Popen creationflags parameter to specify that a new process does
          not inherit the error mode of the calling process. Instead, the
          new process gets the default error mode. This feature is
          particularly useful for multithreaded shell applications that
          run with hard errors disabled.
      subprocess.CREATE_BREAKAWAY_FROM_JOB
          A Popen creationflags parameter to specify that a new process is
          not associated with the job.
    """
    # si = subprocess.STARTUPINFO()
    flags = 0  # subprocess.DETACHED_PROCESS

    process = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        creationflags=flags,
        ## not required when processing by read(1)
        # bufsize=1,
        ## Adding universal newlines will evoke _str_ from the `out`
        # universal_newlines=True,
    )

    byte_count = 0
    count = 0
    while True:
        out = process.stdout.read(1)
        if out == "" and process.poll() != None:
            break
        byte_count += 1

        if byte_count > 75_000:
            break

        if len(out) > 0:
            sys.stdout.write(clean(out, ""))
            sys.stdout.flush()
        else:
            count += 1
            if count % 20 == 0:
                print("blanked.")
                break
    return proc_wait(process)


def proc_wait(proc, timeout=10):
    """Wait for a process to complete and return its output."""
    try:
        outs, errs = proc.communicate(timeout=timeout)
    except subprocess.TimeoutExpired:
        proc.kill()
        outs, errs = proc.communicate()
    return outs, errs


def clean(text, default=None):
    """Clean the text input, ensuring it is a string."""
    t = text
    if isinstance(text, bytes):
        t = text.decode("utf")
    return default if len(t) == 0 else t



import unittest
from unittest.mock import patch

from trim import urls


class TestUrls(unittest.TestCase):
    """Unit tests for URL-related functionality."""

    @patch("trim.urls.reverse")
    @patch("trim.urls.absolutify")
    def test_absolute_reverse(self, mock_absolutify, mock_reverse):
        """Test that absolute_reverse returns the correct absolute URL."""
        # Setup
        mock_reverse.return_value = "/some/path/"
        request = "dummy_request"
        # Expected: Absolute URL should be constructed correctly
        expected_url = "http://example.com/some/path/"
        mock_absolutify.return_value = expected_url

        # Result
        name = ("some_view_name",)
        args = ()
        result = urls.absolute_reverse(request, name, *args)

        # Check if the result matches expected URL
        self.assertEqual(result, expected_url)
        mock_reverse.assert_called_once_with(name, args=args)
        mock_absolutify.assert_called_once_with(request, "/some/path/")

    def test_absolutify(self):
        """Test that absolutify constructs the correct absolute URL."""

        class MockRequest:
            scheme = "http"
            get_host = lambda self: "example.com"

        request = MockRequest()
        relative_url = "/some/path/"
        # Expected: Absolute URL should be constructed correctly
        expected_url = "http://example.com/some/path/"
        # Result
        result = urls.absolutify(request, relative_url)
        # Check if the result matches expected URL
        self.assertEqual(result, expected_url)

    @patch("trim.urls.staticfiles_storage")
    @patch("trim.urls.path")
    def test_favicon_path(self, mock_path, mock_staticfiles_storage):
        """Test that favicon_path returns the correct path."""
        # Expected: Favicon path should be '/static/favicon.ico'
        expected_path = "/static/favicon.ico"
        mock_path.return_value = expected_path
        # Result
        result = urls.favicon_path()
        # Check if the result matches expected path
        self.assertEqual(result, expected_path)

    def test_clean_str_with_none(self):
        """Test clean_str returns empty string for None."""
        self.assertEqual(urls.clean_str(None), "")

    def test_clean_str_with_false(self):
        """Test clean_str returns empty string for False."""
        self.assertEqual(urls.clean_str(False), "")

    def test_clean_str_with_string(self):
        """Test clean_str returns the string unchanged."""
        self.assertEqual(urls.clean_str("hello"), "hello")

    def test_clean_str_with_empty_string(self):
        """Test clean_str returns empty string unchanged."""
        self.assertEqual(urls.clean_str(""), "")

    @patch("trim.urls.paths")
    @patch("trim.urls.django_include")
    def test_path_includes_single_name(self, mock_include, mock_paths):
        """Test path_includes with single string name."""
        # Setup
        mock_paths.return_value = "result"
        # Execute
        result = urls.path_includes("products")
        # Assert
        expected = (("products", ("products/", mock_include.return_value)),)
        mock_include.assert_called_once_with("products.urls")
        mock_paths.assert_called_once_with(expected)
        self.assertEqual(result, "result")

    @patch("trim.urls.paths")
    @patch("trim.urls.django_include")
    def test_path_includes_multiple_names(self, mock_include, mock_paths):
        """Test path_includes with multiple names."""
        # Setup
        mock_paths.return_value = "result"
        # Execute
        result = urls.path_includes("products", "orders")
        # Assert
        self.assertEqual(mock_include.call_count, 2)
        mock_include.assert_any_call("products.urls")
        mock_include.assert_any_call("orders.urls")
        self.assertEqual(result, "result")

    @patch("trim.urls.paths")
    @patch("trim.urls.django_include")
    def test_path_includes_with_nested_tuple(self, mock_include, mock_paths):
        """Test path_includes flattens nested tuples."""
        # Setup
        mock_paths.return_value = "result"
        # Execute
        result = urls.path_includes(("products", "orders"))
        # Assert
        self.assertEqual(mock_include.call_count, 2)
        mock_include.assert_any_call("products.urls")
        mock_include.assert_any_call("orders.urls")
        self.assertEqual(result, "result")

    @patch("trim.urls.paths")
    @patch("trim.urls.django_include")
    def test_path_includes_pair_with_string(self, mock_include, mock_paths):
        """Test path_includes_pair auto-generates URL from string."""
        # Setup
        mock_paths.return_value = "result"
        # Execute
        result = urls.path_includes_pair("products")
        # Assert
        expected = (("products", ("products/", mock_include.return_value)),)
        mock_include.assert_called_once_with("products.urls")
        mock_paths.assert_called_once_with(expected)
        self.assertEqual(result, "result")

    @patch("trim.urls.paths")
    @patch("trim.urls.django_include")
    def test_path_includes_pair_with_custom_url(self, mock_include, mock_paths):
        """Test path_includes_pair uses custom URL from tuple."""
        # Setup
        mock_paths.return_value = "result"
        # Execute
        result = urls.path_includes_pair(("trim.account", "account/"))
        # Assert
        expected = (("trim.account", ("account/", mock_include.return_value)),)
        mock_include.assert_called_once_with("trim.account.urls")
        mock_paths.assert_called_once_with(expected)
        self.assertEqual(result, "result")

    @patch("trim.urls.paths")
    @patch("trim.urls.django_include")
    def test_path_includes_pair_mixed(self, mock_include, mock_paths):
        """Test path_includes_pair with mixed string and tuple."""
        # Setup
        mock_paths.return_value = "result"
        # Execute
        result = urls.path_includes_pair("file", ("trim.account", "account/"))
        # Assert
        self.assertEqual(mock_include.call_count, 2)
        mock_include.assert_any_call("file.urls")
        mock_include.assert_any_call("trim.account.urls")
        self.assertEqual(result, "result")

    @patch("trim.urls.paths")
    @patch("trim.urls.django_include")
    def test_path_include_basic(self, mock_include, mock_paths):
        """Test path_include adds trailing slash to URL."""
        # Setup
        mock_paths.return_value = "result"
        # Execute
        result = urls.path_include("products")
        # Assert
        expected = (("products", ("products/", mock_include.return_value)),)
        mock_include.assert_called_once_with("products.urls")
        mock_paths.assert_called_once_with(expected)
        self.assertEqual(result, "result")

    @patch("trim.urls.paths")
    @patch("trim.urls.django_include")
    def test_path_include_with_custom_module(self, mock_include, mock_paths):
        """Test path_include uses custom module."""
        # Setup
        mock_paths.return_value = "result"
        # Execute
        result = urls.path_include("mythings/", "products.urls")
        # Assert
        expected = (("ythings/", ("mythings/", mock_include.return_value)),)
        mock_include.assert_called_once_with("products.urls")
        mock_paths.assert_called_once_with(expected)
        self.assertEqual(result, "result")

    @patch("trim.urls.paths")
    @patch("trim.urls.django_include")
    def test_path_include_with_custom_name(self, mock_include, mock_paths):
        """Test path_include uses custom path name."""
        # Setup
        mock_paths.return_value = "result"
        # Execute
        result = urls.path_include("mythings/", "products.urls", "items")
        # Assert
        expected = (("items", ("mythings/", mock_include.return_value)),)
        mock_include.assert_called_once_with("products.urls")
        mock_paths.assert_called_once_with(expected)
        self.assertEqual(result, "result")

    @patch("trim.urls.paths_less")
    @patch("trim.urls.trim_names")
    def test_paths_default_builds_patterns(self, mock_trim_names, mock_paths_less):
        """Test paths_default builds patterns dict from views."""
        # Setup
        mock_trim_names.get_url.side_effect = ["", "new/", "<str:pk>/"]
        mock_paths_less.return_value = "result"
        views_list = ("list", "create", "detail")
        # Execute
        result = urls.paths_default("mock_views", "mock_models", views=views_list)
        # Assert
        self.assertEqual(mock_trim_names.get_url.call_count, 3)
        mock_paths_less.assert_called_once_with(
            views="mock_views",
            model_list="mock_models",
            ignore_missing_views=True,
            list="",
            create="new/",
            detail="<str:pk>/",
        )
        self.assertEqual(result, "result")

    @patch("trim.urls.paths_less")
    @patch("trim.urls.trim_names")
    def test_paths_default_passes_ignore_flag(self, mock_trim_names, mock_paths_less):
        """Test paths_default passes ignore_missing_views flag."""
        # Setup
        mock_trim_names.get_url.return_value = ""
        mock_paths_less.return_value = "result"
        # Execute
        urls.paths_default(
            "views", "models", ignore_missing_views=False, views=("list",)
        )
        # Assert
        mock_paths_less.assert_called_once_with(
            views="views", model_list="models", ignore_missing_views=False, list=""
        )

    @patch("trim.urls.paths_named")
    def test_paths_less_converts_single_model(self, mock_paths_named):
        """Test paths_less converts single model to tuple."""
        # Setup
        mock_model = type("Product", (), {"__name__": "Product"})()
        mock_paths_named.return_value = ["path1", "path2"]
        # Execute
        result = urls.paths_less("views", mock_model, list="", create="new/")
        # Assert
        mock_paths_named.assert_called_once_with(
            "views",
            "Product",
            url_pattern_prefix="product/",
            ignore_missing_views=False,
            url_name_prefix="product-",
            list="",
            create="new/",
        )
        self.assertEqual(result, ["path1", "path2"])

    @patch("trim.urls.paths_named")
    def test_paths_less_handles_multiple_models(self, mock_paths_named):
        """Test paths_less calls paths_named for each model."""
        # Setup
        mock_model1 = type("Product", (), {"__name__": "Product"})()
        mock_model2 = type("Order", (), {"__name__": "Order"})()
        mock_paths_named.side_effect = [["path1"], ["path2"]]
        # Execute
        result = urls.paths_less("views", (mock_model1, mock_model2), list="")
        # Assert
        self.assertEqual(mock_paths_named.call_count, 2)
        self.assertEqual(result, ["path1", "path2"])

    @patch("trim.urls.paths_named")
    def test_paths_less_passes_ignore_flag(self, mock_paths_named):
        """Test paths_less passes ignore_missing_views flag."""
        # Setup
        mock_model = type("Product", (), {"__name__": "Product"})()
        mock_paths_named.return_value = []
        # Execute
        urls.paths_less("views", mock_model, ignore_missing_views=True)
        # Assert
        call_args = mock_paths_named.call_args[1]
        self.assertTrue(call_args["ignore_missing_views"])

    @patch("trim.urls.paths_dict")
    @patch("trim.urls.MAPPED_NAMES", {"list": "ListView"})
    def test_paths_named_converts_string_to_tuple(self, mock_paths_dict):
        """Test paths_named converts string pattern to tuple using MAPPED_NAMES."""
        # Setup
        mock_paths_dict.return_value = "result"
        # Execute
        result = urls.paths_named("views", list="")
        # Assert
        expected_patterns = {"ListView": ("list", "")}
        mock_paths_dict.assert_called_once_with(
            "views",
            expected_patterns,
            None,
            url_pattern_prefix=None,
            url_name_prefix=None,
            ignore_missing_views=False,
        )
        self.assertEqual(result, "result")

    @patch("trim.urls.paths_dict")
    def test_paths_named_preserves_tuple_patterns(self, mock_paths_dict):
        """Test paths_named preserves already-tuple patterns."""
        # Setup
        mock_paths_dict.return_value = "result"
        # Execute
        urls.paths_named("views", list=("ListView", ""))
        # Assert
        call_args = mock_paths_dict.call_args[0]
        patterns = call_args[1]
        self.assertEqual(patterns["ListView"], ("list", ""))

    @patch("trim.urls.paths_dict")
    def test_paths_named_passes_all_params(self, mock_paths_dict):
        """Test paths_named forwards all parameters to paths_dict."""
        # Setup
        mock_paths_dict.return_value = "result"
        # Execute
        urls.paths_named(
            "views",
            view_prefix="Product",
            ignore_missing_views=True,
            url_pattern_prefix="p/",
            url_name_prefix="prod-",
        )
        # Assert
        mock_paths_dict.assert_called_once_with(
            "views",
            {},
            "Product",
            url_pattern_prefix="p/",
            url_name_prefix="prod-",
            ignore_missing_views=True,
        )

    @patch("trim.urls.paths_dict")
    def test_paths_tuple_converts_tuples_to_dict(self, mock_paths_dict):
        """Test paths_tuple converts tuple patterns to dict format."""
        # Setup
        mock_paths_dict.return_value = "result"
        patterns = (("ListView", "list", ""), ("CreateView", "create", "new/"))
        # Execute
        result = urls.paths_tuple("views", patterns)
        # Assert
        expected_dict = {"ListView": ["list", ""], "CreateView": ["create", "new/"]}
        mock_paths_dict.assert_called_once_with("views", expected_dict)
        self.assertEqual(result, "result")

    @patch("trim.urls.paths_dict")
    def test_paths_tuple_forwards_kwargs(self, mock_paths_dict):
        """Test paths_tuple forwards additional kwargs to paths_dict."""
        # Setup
        mock_paths_dict.return_value = "result"
        patterns = (("ListView", "list", ""),)
        # Execute
        urls.paths_tuple("views", patterns, view_prefix="Product")
        # Assert
        call_kwargs = mock_paths_dict.call_args[1]
        self.assertEqual(call_kwargs["view_prefix"], "Product")



from django.contrib.auth import views as auth_views
from django.urls import include, path, re_path

from trim import urls as t_urls

from . import views

app_name = "account"

urlpatterns = t_urls.paths_named(
    views,
    profile=(
        "ProfileView",
        (
            "",
            "profile/",
        ),
    ),
    edit=(
        "ProfileUpdateView",
        "profile/edit/",
    ),
    edit_username=(
        "ProfileUsernameUpdateView",
        "profile/edit/username/",
    ),
    edit_password=(
        "ProfilePasswordUpdateView",
        "profile/edit/user-information/",
    ),
    edit_email=(
        "ProfileEmailUpdateView",
        "profile/edit/email/now/",
    ),
    edit_email_verified=(
        "VerifiedEmailUpdateView",
        "profile/edit/email/",
    ),
    # home=('ProfileView', '',),
    logout=(
        "ProfileLogout",
        "logout/",
    ),
    login=(
        "ProfileLogin",
        "login/",
    ),
    new=(
        "ProfileNewAccount",
        "welcome/",
    ),
    inactive=(
        "ProfileInactiveAccount",
        "inactive/",
    ),
    forgot_success=(
        "ProfileForgotPasswordSuccessView",
        "forgot/sent/",
    ),
    invite=(
        "EmailInviteCreateView",
        "invite/",
    ),
    invites=(
        "EmailInviteListView",
        "invites/",
    ),
)


urlpatterns += [
    path(
        "forgot/",
        views.PasswordResetView.as_view(),
        name="admin_password_reset",
    ),
    path(
        "profile/edit/email/verify/",
        views.VerifyEmailTokenView.as_view(),
        name="verify_email_token",
    ),
    path(
        "profile/edit/email/verify/<str:uuid>/",
        views.VerifyEmailTokenView.as_view(),
        name="verify_email_token",
    ),
]



from django.template.base import Template
from django.template.response import TemplateResponse

try:
    import markdown as markdown_orig
except ImportError:
    markdown_orig = None


__all__ = [
    "MissingImportError",
    "MarkdownToMarkdownTemplateResponse",
    "MarkdownTemplateResponse",
    "MarkdownDoubleTemplateResponse",
    "MarkdownReponseMixin",
]


class MissingImportError(ImportError):
    """The trim.markdown.MarkdownTemplateResponse required a markdown
    renderer. If missing (at execution time), raise this import error.
    """

    pass


class MarkdownTemplateResponse(TemplateResponse):
    """Expecting a .md template file - perform standard rendering, coverting
    to HTML and return the finished page.
    """

    @property
    def rendered_content(self):
        """Return the freshly rendered content for the template and context
        described by the TemplateResponse.

        This *does not* set the final content of the response. To set the
        response content, you must either call render(), or set the
        content explicitly using the value of this property.
        """
        # django.template.backends.django.Template
        template = self.resolve_template(self.template_name)
        context = self.resolve_context(self.context_data)

        source_text = template.template.source
        md = self.get_markdown_object(context)
        html = md.convert(source_text)

        meta = md.Meta
        context["metadata"] = meta

        old_inner = template.template
        inner = Template(html, old_inner.origin, old_inner.name, old_inner.engine)
        template.template = inner
        res = template.render(context, self._request)
        return res

    def get_markdown_object(self, context):
        # context['view']

        if "view" in context:
            if hasattr(context["view"], "get_markdown_object"):
                return context["view"].get_markdown_object()
        # meta into the context.
        # HTML is the raw
        # https://python-markdown.github.io/extensions/
        extensions = [
            "meta",
            "extra",
        ]

        if markdown_orig is None:
            raise MissingImportError("markdown module is not installed.")

        md = markdown_orig.Markdown(extensions=extensions)
        return md


class MarkdownToMarkdownTemplateResponse(MarkdownTemplateResponse):

    @property
    def rendered_content(self):
        """Return the freshly rendered content for the template and context
        described by the TemplateResponse.

        This *does not* set the final content of the response. To set the
        response content, you must either call render(), or set the
        content explicitly using the value of this property.
        """
        # django.template.backends.django.Template
        template = self.resolve_template(self.template_name)
        context = self.resolve_context(self.context_data)

        source_text = template.template.source
        old_inner = template.template
        inner = Template(
            source_text, old_inner.origin, old_inner.name, old_inner.engine
        )
        template.template = inner
        res = template.render(context, self._request)
        return res


class MarkdownDoubleTemplateResponse(MarkdownTemplateResponse):
    """USe the default template as the wrapper for the markdown -
    the markdown has its own render chain.
    Provide the markdown object to the context of the original template
    """

    @property
    def rendered_content(self):
        """Return the freshly rendered content for the template and context
        described by the TemplateResponse.

        This *does not* set the final content of the response. To set the
        response content, you must either call render(), or set the
        content explicitly using the value of this property.
        """
        # django.template.backends.django.Template
        context = self.resolve_context(self.context_data)

        view_template = self.resolve_template(self.template_name)
        view_source = view_template.template.source
        view_inner = view_template.template

        # Convert the view source (the template_name markdown file) to HTML
        md = self.get_markdown_object(context)
        markdown_html = md.convert(view_source)
        # The markdown template is considered incomplete view template
        md_inner = Template(
            markdown_html, view_inner.origin, view_inner.name, view_inner.engine
        )
        view_template.template = md_inner
        meta = md.Meta

        context["metadata"] = meta
        context["markdown"] = {
            # render the incomplete markdown HTML to finished HTML
            "html": view_template.render(context, self._request),
            "object": md,
            "meta": meta,
        }

        # grab the template name from the meta data within the markdown.
        layout_template = self.resolve_template(meta["template_name"])
        # The layout (the template_name from the meta of the original view markdown file)
        return layout_template.render(context, self._request)


class MarkdownReponseMixin(object):
    response_class = MarkdownTemplateResponse
    # response_class = MarkdownDoubleTemplateResponse



"""
The "live" models tool provides a fast access method to all the existing
installed models using a single concat string, resolving to the model.

This can replace a standard model import, or the django get_model import:


with get_model:

    from django.apps import apps

    ShippingAccount = apps.get_model('baskets.ShippingAccount')
    cart_owner = self.get_object().cart.owner
    user_addresses = ShippingAccount.objects.filter(owner=cart_owner, deleted=False)

after:

    from trim import live

    owner = self.get_object().cart.owner
    user_addresses = live.baskets.ShippingAccount.objects.filter(
            owner=owner, deleted=False
        )
"""

from django.apps import apps


class MagicModelApp(object):
    """The top level 'live' model import, expecting a string caller:

    live = MagicModelApp()
    live.appname

    """

    def __getattribute__(self, k):
        # Use __getattribute__ to intercept all attribute access, not just missing ones
        # This allows us to properly raise AttributeError for special attributes
        # before __getattr__ is even called
        if k.startswith("_") or k in (
            "__class__",
            "__dict__",
            "__wrapped__",
            "__name__",
            "__qualname__",
        ):
            # Delegate to parent for special attributes
            return object.__getattribute__(self, k)
        # Return a MagicModelModel for app name lookups
        return MagicModelModel(k)


class MagicModelModel(object):
    """A return value from the live appname collection, returning
    a model upon a key call:

        live.appname.ModelName
        <appname.ModelName>
    """

    def __init__(self, appname):
        object.__setattr__(self, "appname", appname)

    def __getattribute__(self, k):
        # Use __getattribute__ to intercept all attribute access
        if k.startswith("_") or k in ("appname",):
            # Allow access to our internal appname attribute
            # Delegate to parent for special attributes
            return object.__getattribute__(self, k)
        # Return the model from Django's registry
        appname = object.__getattribute__(self, "appname")
        return apps.get_model(appname, k)


# class MagicModelProps(object):

#     def __init__(self, appname, modelname):
#         self.appname = appname
#         self.modelname = modelname

#     def __getattr__(self, k):

live = MagicModelApp()



import textwrap

from django import template
from django.conf import settings
from django.urls import resolve, reverse

from .shared_tools import parse_tag, parse_until

try:
    import markdown as markdown_orig
except ImportError:
    markdown_orig = None


register = template.Library()


@register.tag(name="markdown")
def do_slot(parser, token):
    nodelist, splits, extra = parse_until(parser, token, ("endmarkdown",))
    return MarkdownContentNode(nodelist, splits, **extra)


class MarkdownContentNode(template.Node):

    def __init__(self, nodelist, tokens, *a, **kw):
        self.nodelist = nodelist

        self.token_template_name = None
        if len(tokens) > 0:
            self.token_template_name = template.Variable(tokens[0])
        self.extra_context = kw

    def render(self, context):
        print("rendering markdown")

        wrap = {}
        md = get_markdown_object(context, **self.extra_context)

        # Make a resolved dict of values given through the init.
        values = {key: val.resolve(context) for key, val in self.extra_context.items()}
        with context.push(**wrap, **values):
            # Context({'source': content}, autoescape=context.autoescape)
            ## This is markdown plain - convert to rendered markdown
            django_markdown_text = self.nodelist.render(context)
            # django_markdown_text = '\n'.join([m.lstrip() for m in django_markdown_text.split('\n')])

            if django_markdown_text[0] == "\n":
                django_markdown_text = django_markdown_text[1:]

            plain_markdown_text = textwrap.dedent(django_markdown_text)
            return md.convert(plain_markdown_text)


from django.template.loader_tags import (
    Node,
    TemplateSyntaxError,
    construct_relative_path,
    token_kwargs,
)


@register.tag("markdown.file_2")
def do_markdown_file(parser, token):
    """
    Load a template and render it with the current context. You can pass
    additional context using keyword arguments.

    _This is a direct clone of the include node with an alterntive Node parser._

    Example::

        {% markdown.file "foo" "some/include.md" %}
        {% markdown.file "foo" "some/include.md" with bar="BAZZ!" baz="BING!" %}

    Use the ``only`` argument to exclude the current context when rendering
    the included template::

        {% markdown.file "foo" "some/include.md" only %}
        {% markdown.file "foo" "some/include.md" with bar="1" only %}
    """
    # bits = token.split_contents()
    # if len(bits) < 2:
    #     raise TemplateSyntaxError(
    #         "%r tag takes at least one argument: the name of the template to "
    #         "be included." % bits[0]
    #     )

    # options = {}
    # remaining_bits = bits[2:]
    # while remaining_bits:
    #     option = remaining_bits.pop(0)
    #     if option in options:
    #         raise TemplateSyntaxError(
    #             "The %r option was specified more than once." % option
    #         )
    #     if option == "with":
    #         value = token_kwargs(remaining_bits, parser, support_legacy=False)
    #         if not value:
    #             raise TemplateSyntaxError(
    #                 '"with" in %r tag needs at least one keyword argument.' % bits[0]
    #             )
    #     elif option == "only":
    #         value = True
    #     else:
    #         raise TemplateSyntaxError(
    #             "Unknown argument for %r tag: %r." % (bits[0], option)
    #         )
    #     options[option] = value

    # namemap = options.get("with", {})

    bits, namemap, options = parse_tag(parser, token)

    # bits[1] = construct_relative_path(parser.origin.template_name, bits[1])
    isolated_context = options.get("only", False)

    c_bits = tuple(parser.compile_filter(value) for value in bits)
    return IncludeNode(
        c_bits,
        extra_context=namemap,
        isolated_context=isolated_context,
    )


class IncludeNode(Node):
    context_key = "__include_context"

    def __init__(
        self, template_bits, *args, extra_context=None, isolated_context=False, **kwargs
    ):
        self.template_bits = template_bits
        self.extra_context = extra_context or {}
        self.isolated_context = isolated_context
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f"<{self.__class__.__qualname__}: template_bits={self.template_bits!r}>"

    def render(self, context):
        """
        Render the specified template and context. Cache the template object
        in render_context to avoid reparsing and loading when used in a for
        loop.
        """

        # convert the vars into resolved strings
        resolved_bits = (x.resolve(context) for x in self.template_bits)

        ## In the original version, this _template_ (resolved str) is a
        # path to a template. However here we have content to resolve.
        # (like a block...)

        # engine.from_string(...)

        # template = self.template.resolve(context)
        # Does this quack like a Template?
        if not callable(getattr(template, "render", None)):
            # If not, try the cache and select_template().
            template_name = template or ()
            if isinstance(template_name, str):
                template_name = (
                    construct_relative_path(
                        self.origin.template_name,
                        template_name,
                    ),
                )
            else:
                template_name = tuple(template_name)
            cache = context.render_context.dicts[0].setdefault(self, {})
            template = cache.get(template_name)
            if template is None:
                template = context.template.engine.select_template(template_name)
                cache[template_name] = template
        # Use the base.Template of a backends.django.Template.
        elif hasattr(template, "template"):
            template = template.template

        values = {
            name: var.resolve(context) for name, var in self.extra_context.items()
        }

        # info = get_file_contents(bits[1], bits[0])

        if self.isolated_context:
            return template.render(context.new(values))
        with context.push(**values):
            return template.render(context)


@register.inclusion_tag(
    "trim/templatetags/markdown_file_content.html",
    takes_context=True,
    name="markdown.file",
)
def src_code_content_template(context, part_a=None, part_b=None, *args, **kwargs):
    """Load and process a markdown file.

    The target file may be an _absolute_ path or _relative_ within a
    TRIM_MARKDOWN_DIRS entry.

    absolute:

        {% markdown.file '/abs/path/to/readme.md' %}

    relative takes two parts:

        {% markdown.file "docs" 'points/readme.md' %}

    The `base_dir` `"docs"` is populated give the settings config:

            # a pack of locations we can use for {% markdown.file filepath %}
            TRIM_MARKDOWN_DIRS = {
                'point_src': POLYPOINT_SRC_DIR,
                'point_src_rel': POLYPOINT_THEATRE_SRC_RELATIVE_PATH,
                'site_demos': POLYPOINT_DEMO_DIR,
                'docs': POLYPOINT_DOCS_DIR,
            }

    The `TRIM_MARKDOWN_DIRS` entry (such as "docs") should be an absolute path,
    allowing the file target `[base_dir] / [relative_path]`.

    """
    base_dir = None
    filename = part_a

    if part_b is not None:
        base_dir = settings.TRIM_MARKDOWN_DIRS.get(part_a)
        # dir, filename.
        filename = part_b

    info = get_file_contents(filename, base_dir)
    if info["exists"]:
        content = info["content"]
        md = get_markdown_object(context, **kwargs)
        info["html"] = md.convert(content)
        info["metadata"] = md.Meta
    else:
        print("Does not exist", info)

    return {
        "markdown_object": info,
    }


@register.inclusion_tag(
    "trim/templatetags/markdown_file_content.html",
    takes_context=True,
    name="markdown.text",
)
def src_code_content_text(context, *args, **kwargs):
    """Load and process markdown from a var or string.

    {% markdown.text 'This _content_ is rendered through **Markdown**' %}
    """

    content = "\n".join(args)

    info = {
        "filepath": None,
        "exists": True,
        "content": content,
    }

    md = get_markdown_object(context, **kwargs)

    if content[0] == "\n":
        content = content[1:]

    plain_markdown_text = textwrap.dedent(content)
    info["html"] = md.convert(plain_markdown_text)
    info["metadata"] = md.Meta

    return {
        "markdown_object": info,
    }


def get_markdown_object(context, **kwargs):
    # context['view']

    md = kwargs.get("markdown_object")
    if md is None:
        md = context.get("get_markdown_object")
        if hasattr(context.get("view", {}), "get_markdown_object"):
            md = context["view"].get_markdown_object()

    if md:
        return md
    # meta into the context.
    # HTML is the raw
    # https://python-markdown.github.io/extensions/
    extensions = [
        "meta",
        # 'extra',
        "pymdownx.extra",
    ]

    if markdown_orig is None:
        raise MissingImportError("markdown module is not installed.")

    md = markdown_orig.Markdown(extensions=extensions)
    return md


from pathlib import Path


def get_file_contents(path, parent=None, safe=True):
    if parent is None and safe is True:
        target = Path(".") / path
        exists = target.exists()
        return {
            "filepath": target,
            "exists": exists,
            "error": "no parent defined",
        }
    target = Path(parent) / path
    print("target", target)
    exists = target.exists()
    content = None
    if exists and target.is_file():
        content = target.read_text()
    else:
        exists = False
    return {
        "filepath": target,
        "exists": exists,
        "content": content,
    }



"""
Test the trim.tests module itself to ensure testing utilities work correctly.
"""

import pytest
from django.contrib.auth.mixins import LoginRequiredMixin, PermissionRequiredMixin
from django.views.generic import DetailView, ListView, UpdateView

from trim.tests import (
    MixinOrderError,
    assert_mixin_order,
    assert_view_has_permission,
    assert_view_requires_login,
    assert_views_mixin_order,
    get_view_permissions,
)


# Test views with correct order
class CorrectView(LoginRequiredMixin, UpdateView):
    """Properly ordered view"""

    pass


class CorrectPermissionView(PermissionRequiredMixin, ListView):
    """Properly ordered view with permissions"""

    permission_required = "myapp.view_model"


# Test views with incorrect order
class IncorrectView(UpdateView, LoginRequiredMixin):
    """Incorrectly ordered view - mixin after view"""

    pass


class IncorrectPermissionView(DetailView, PermissionRequiredMixin):
    """Incorrectly ordered view - mixin after view"""

    permission_required = "myapp.change_model"


# Test views with no auth mixins
class PlainView(ListView):
    """View without any auth mixins"""

    pass


class TestMixinOrderValidation:
    """Test mixin order validation functions"""

    def test_correct_order_passes(self):
        """Correctly ordered views should pass validation"""
        is_valid, error = assert_mixin_order(CorrectView, raise_exception=False)
        assert is_valid
        assert error is None

    def test_correct_permission_view_passes(self):
        """Correctly ordered permission views should pass"""
        is_valid, error = assert_mixin_order(
            CorrectPermissionView, raise_exception=False
        )
        assert is_valid
        assert error is None

    def test_incorrect_order_fails(self):
        """Incorrectly ordered views should fail validation"""
        with pytest.raises(MixinOrderError) as exc_info:
            assert_mixin_order(IncorrectView)

        error_msg = str(exc_info.value)
        assert "IncorrectView" in error_msg
        assert "LoginRequiredMixin" in error_msg
        assert "UpdateView" in error_msg

    def test_incorrect_order_no_exception(self):
        """Test non-raising mode returns error info"""
        is_valid, error = assert_mixin_order(IncorrectView, raise_exception=False)
        assert not is_valid
        assert error is not None
        assert "IncorrectView" in error

    def test_plain_view_passes(self):
        """Views without auth mixins should pass (nothing to validate)"""
        is_valid, error = assert_mixin_order(PlainView, raise_exception=False)
        assert is_valid
        assert error is None


class TestModuleValidation:
    """Test module-wide view validation"""

    def test_validate_correct_module(self):
        """Module with only correct views should pass"""
        import types

        # Create a mock module with correct views
        mock_module = types.ModuleType("mock_views")
        mock_module.CorrectView = CorrectView
        mock_module.CorrectPermissionView = CorrectPermissionView

        errors = assert_views_mixin_order(mock_module, raise_exception=False)
        assert len(errors) == 0

    def test_validate_incorrect_module(self):
        """Module with incorrect views should fail"""
        import types

        mock_module = types.ModuleType("mock_views")
        mock_module.IncorrectView = IncorrectView
        mock_module.IncorrectPermissionView = IncorrectPermissionView

        errors = assert_views_mixin_order(mock_module, raise_exception=False)
        assert len(errors) == 2

    def test_validate_mixed_module_raises(self):
        """Module with any incorrect view should raise"""
        import types

        mock_module = types.ModuleType("mock_views")
        mock_module.CorrectView = CorrectView
        mock_module.IncorrectView = IncorrectView

        with pytest.raises(MixinOrderError):
            assert_views_mixin_order(mock_module, raise_exception=True)


class TestPermissionInspection:
    """Test permission inspection utilities"""

    def test_get_permissions_from_view(self):
        """Should extract permissions from view class"""
        info = get_view_permissions(CorrectPermissionView)

        assert info["permissions"] == ["myapp.view_model"]
        assert info["login_required"] is False
        assert info["any"] is False

    def test_get_login_required(self):
        """Should detect login requirement"""
        info = get_view_permissions(CorrectView)

        assert info["login_required"] is True

    def test_plain_view_has_no_requirements(self):
        """Plain views should have empty requirements"""
        info = get_view_permissions(PlainView)

        assert info["permissions"] == []
        assert info["login_required"] is False
        assert info["user_owned"] is False

    def test_assert_has_permission_passes(self):
        """Should pass when permission exists"""
        assert_view_has_permission(CorrectPermissionView, "myapp.view_model")

    def test_assert_has_permission_fails(self):
        """Should fail when permission missing"""
        with pytest.raises(AssertionError) as exc_info:
            assert_view_has_permission(CorrectPermissionView, "myapp.delete_model")

        assert "does not require permission" in str(exc_info.value)

    def test_assert_requires_login_passes(self):
        """Should pass when login required"""
        assert_view_requires_login(CorrectView)

    def test_assert_requires_login_fails(self):
        """Should fail when login not required"""
        with pytest.raises(AssertionError) as exc_info:
            assert_view_requires_login(PlainView)

        assert "does not require login" in str(exc_info.value)


class TestMultiplePermissions:
    """Test views with multiple permissions"""

    def test_multiple_permissions_list(self):
        """Should handle list of permissions"""

        class MultiPermView(PermissionRequiredMixin, ListView):
            permission_required = ["myapp.view_model", "myapp.change_model"]

        info = get_view_permissions(MultiPermView)
        assert len(info["permissions"]) == 2
        assert "myapp.view_model" in info["permissions"]
        assert "myapp.change_model" in info["permissions"]

    def test_single_permission_string(self):
        """Should handle single permission as string"""
        info = get_view_permissions(CorrectPermissionView)
        assert info["permissions"] == ["myapp.view_model"]



import os
from pathlib import Path


def recombine(dir_path, output_filepath=None):
    """
    Merge multiple split file parts from a directory into a single file.

    This function takes a directory containing split file parts (typically with
    extensions like .part_0, .part_1, etc.) and recombines them into a single
    output file. The parts are merged in numerical order based on their suffix.

    Args:
        dir_path (str or Path): The directory path containing the split file parts
            to be recombined.
        output_filepath (str or Path, optional): The path where the merged file
            should be written. Can be either a file path or a directory path.
            If a directory is provided, the output filename will be derived from
            the first part file (without the .part_N extension). If None, defaults
            to the input directory. Defaults to None.

    Returns:
        Path: The Path object pointing to the created merged file.

    Raises:
        FileExists: If the output file already exists at the specified location.

    Examples:
        Merge parts in a directory to the same directory:

            >>> output = recombine('/path/to/parts/')
            >>> # Creates merged file in /path/to/parts/filename

        Specify a custom output file path:

            >>> output = recombine('/path/to/parts/', '/output/merged.zip')
            >>> print(output)
            /output/merged.zip

        Specify an output directory:

            >>> output = recombine('/path/to/parts/', '/output/dir/')
            >>> # Creates merged file in /output/dir/filename

    Notes:
        - Part files must have extensions in the format .part_0, .part_1, etc.
        - Parts are sorted numerically before merging
        - The function prints progress information during the merge process
        - The output file size is printed upon completion
        - If output file exists, FileExists exception is raised to prevent overwriting

    See Also:
        split_i: Helper function used to extract part numbers for sorting
    """

    # get all files
    dir_files = os.listdir(dir_path)
    print("Found", len(dir_files))

    newname = Path(dir_files[0]).stem  # drops the .part_0

    # same DIR, file is the first part.
    output_filepath = Path(output_filepath or dir_path)

    if output_filepath.is_dir():
        # Apply the name
        output_filepath = output_filepath / newname

    print("output_filepath", output_filepath)

    if output_filepath.exists():
        print("Output file exists", output_filepath)
        raise FileExists(output_filepath)

    ordered_files = sorted(dir_files, key=split_i)
    # now merge
    with open(output_filepath, "wb+") as write_stream:
        print("Writing", output_filepath)
        for filename in ordered_files:
            subfile = Path(dir_path) / filename
            print("  ", filename)
            with open(subfile, "rb") as read_stream:
                write_stream.write(read_stream.read())
        print("Complete", output_filepath.name)

    print("Size: ", os.path.getsize(output_filepath))

    return output_filepath


def split_i(item):
    return int(Path(item).suffix.split("_")[1])


class FileExists(Exception):
    pass



"""
Short URL to help lighten the load for dev urls

    default:

        from django.urls import path

        urlpatterns = [
            path('', views.ProductListView.as_view(), name='list'),
            path('new/', views.ProductCreateView.as_view(), name='create'),
            path('change/<str:pk>/', views.ProductUpdateView.as_view(), name='update'),
            path('delete/<str:pk>/', views.ProductDeleteView.as_view(), name='delete'),
            path('<str:pk>/', views.ProductDetailView.as_view(), name='detail'),
        ]


    to this:

        from . import models
        from trim.models import grab_models

        urlpatterns = trims.paths_default(views, grab_models(models),
            views=('list', 'create', 'update', 'delete', 'detail',),
        )


    or this:

        from . import models
        from trim.models import grab_models

        urlpatterns = trims.paths_less(views, grab_models(models),
            ignore_missing_views=True,
            list='',
            create='new/',
            update='change/<str:pk>/',
            delete='delete/<str:pk>/',
            detail='<str:pk>/',
        )


"""

import inspect
import sys
from functools import reduce

from django.contrib import admin
from django.contrib.staticfiles.storage import staticfiles_storage
from django.urls import include as django_include
from django.urls import path, reverse
from django.views.generic import TemplateView, View
from django.views.generic.base import RedirectView

from trim.names import *

from . import names as trim_names


def absolute_reverse(request, name, *args):
    fa_url = reverse(name, args=args)
    return absolutify(request, fa_url)


def absolutify(request, path):
    base_url = "{0}://{1}{2}".format(
        request.scheme,
        request.get_host(),
        path,
    )
    return base_url


def favicon_path(ingress_path="favicon.ico", static_path="images/{ingress_path}"):
    """Implement your favicon as a static redirect

    # (primary) urls.py
    from trim.urls import favicon_path

    urlpatterns = [
        ...
        favicon_path('favicon.ico'),
    ]
    """
    # Debug prints/breakpoints removed for test stability
    static_path = static_path.format(ingress_path=ingress_path)
    return static_redirect_path(ingress_path, static_path)


def static_redirect_path(ingress_path, static_path):
    return path(
        ingress_path, RedirectView.as_view(url=staticfiles_storage.url(static_path))
    )


urlpatterns = [
    # path('admin/', admin.site.urls),
]


def clean_str(variant):

    if variant in (
        False,
        None,
    ):
        return ""
    return variant


def path_includes(*names):
    """

    from trim.urls import path_includes_pair as includes

    urlpatterns = [
        path("django-admin/", admin.site.urls),
    ] + includes(
            'file',
            'trim.account',
        )
    """
    flat_names = ()
    r = ()

    for name in names:
        if isinstance(name, (list, tuple)) is False:
            name = (name,)
        flat_names += name

    for fln in flat_names:
        d = (
            fln,
            (
                f"{fln}/",
                django_include(f"{fln}.urls"),
            ),
        )
        r += (d,)

    return paths(r)


def path_includes_pair(*items):
    """Similar to the `path_includes` function; given many names, convert to
    a django include and return paths. With `path_includes_pair`, a single item
    may be a list or tuple, allowing ('module', 'url/') expansion.


        from trim.urls import path_includes_pair as includes

        urlpatterns = [
            path("django-admin/", admin.site.urls),
        ] + includes(
                'file',
                ('trim.account', 'account/',),
            )

    """
    pairs = ()
    result = ()

    for item in items:
        if isinstance(item, (list, tuple)) is False:
            # module, url.
            item = (
                item,
                f"{item}/",
            )
            pairs += (item,)
            continue
        pairs += (item,)

    for module, url in pairs:
        entry = (
            module,
            (
                url,
                django_include(f"{module}.urls"),
            ),
        )
        result += (entry,)

    return paths(result)


include = path_includes


def path_urls(views, path_rels):
    for url, view_call in path_rels:
        pass


def index(name):
    return


def path_include(url_name, url_module=None, path_name=None):
    """
        path_include('products/',)
        path_include('mythings/', 'products.urls', ) # name=mythings
        path_include('mythings/', 'products.urls', 'items') # name=items
        path_include('products/', 'products.urls', 'products')

    Apply like this:

        from django.contrib import admin
        from django.urls import path, include

        from trim.urls import path_include, path_includes, error_handlers

        app_name = 'shoppinglist'

        urlpatterns = [
            path('admin/', admin.site.urls),
        ] + path_includes('products')

        error_handlers(__name__)
    """
    r = ()
    url = url_name
    if not url.endswith("/"):
        url = f"{url}/"

    name = path_name or url_name
    if name.endswith("/"):
        name = name[1:]

    urlm = url_module or f"{name}.urls"
    item = (
        name,
        (
            url,
            django_include(urlm),
        ),
    )
    r += (item,)

    return paths(r)


def paths_default(
    views_module, model_list, ignore_missing_views=True, views=None, **options
):
    """
    from . import models
    from trim.models import grab_models

    urlpatterns = trims.paths_default(views, grab_models(models),
        ignore_missing_views=True,
        views=('list', 'create', 'update', 'delete', 'detail',),
    )
    """
    patterns = {}

    for name in views:
        url = trim_names.get_url(name)
        patterns[name] = url

    return paths_less(
        views=views_module,
        model_list=model_list,
        ignore_missing_views=ignore_missing_views,
        **patterns,
    )


def paths_less(views, model_list, ignore_missing_views=False, **patterns):
    """
    from . import models
    from trim.models import grab_models

    urlpatterns = trims.paths_less(views, grab_models(models),
        ignore_missing_views=True,
        list='',
        create='new/',
        update='change/<str:pk>/',
        delete='delete/<str:pk>/',
        detail='<str:pk>/',
    )

    """
    if (
        isinstance(
            model_list,
            (
                list,
                tuple,
            ),
        )
        is False
    ):
        model_list = (model_list,)

    r = []
    for m in model_list:
        name = m.__name__
        unp = name.lower()
        url_pattern_prefix = f"{unp}/"
        r += paths_named(
            views,
            name,
            url_pattern_prefix=url_pattern_prefix,
            ignore_missing_views=ignore_missing_views,
            url_name_prefix=f"{unp}-",
            **patterns,
        )
    return r


def paths_named(
    views,
    view_prefix=None,
    ignore_missing_views=False,
    url_pattern_prefix=None,
    url_name_prefix=None,
    **patterns,
):
    """
        urlpatterns = trims.paths_named(views,
            view_prefix='Product',

            list=('ListView',''),
            create=('CreateView','new/'),
            update=('UpdateView','change/<str:pk>/'),
            delete=('DeleteView','delete/<str:pk>/'),
            detail=('DetailView','<str:pk>/'),
        )

    or less:

        from trim import urls as trims

        urlpatterns = trims.paths_named(views,
            view_prefix='Product',
            # name_genfix="{view_prefix}{mapped_name}View",
            list='',
            create='new/',
            update='change/<str:pk>/',
            delete='delete/<str:pk>/',
            detail='<str:pk>/',
        )
    """
    new_patterns = {}
    for app_name, solution in patterns.items():
        if (
            isinstance(
                solution,
                (
                    list,
                    tuple,
                ),
            )
            is False
        ):
            # only a URL given.
            # convert the 'list' to 'ListView'
            solution = (MAPPED_NAMES.get(app_name), solution)

        class_name, url, *extra = solution
        new_patterns[class_name] = (
            app_name,
            url,
        ) + tuple(extra)

    return paths_dict(
        views,
        new_patterns,
        view_prefix,
        url_pattern_prefix=url_pattern_prefix,
        url_name_prefix=url_name_prefix,
        ignore_missing_views=ignore_missing_views,
    )


def paths_tuple(views, patterns, **kw):
    """Setup URLS using a tuple of tuples, with each given tuple prepared
    and passed to the paths_dict:

        from trim import urls as trims

        trim_patterns = (
            ('NoteIndexView', 'index', '' ),
            ('EntryJsonList', 'entity-list-json', 'entry/list/json/' ),
            ('EntryDetailView', 'entry-detail-json', '<str:pk>/json/' ),
        )

        urlpatterns = trims.paths_tuple(views, trim_patterns)

    Of which is the same as:

        from trim import urls as trims

        trim_patterns = dict(
            NoteIndexView=('index', ''),
            EntryJsonList=('entity-list-json', 'entry/list/json/',),
            EntryDetailView=('entry-detail-json', '<str:pk>/json/', ),
        )

        urlpatterns = trims.paths_dict(views, trim_patterns)
    """
    d_patt = {}

    for key, *params in patterns:
        d_patt[key] = params

    return paths_dict(views, d_patt, **kw)


def paths_dict(
    views,
    patterns=None,
    view_prefix=None,
    ignore_missing_views=False,
    url_pattern_prefix=None,
    url_name_prefix=None,
    safe_prefix=False,
    **kw,
):
    """Given the views module and the patterns,
    generate a short path list.

        trim_patterns = {
            'ProductListView': '',
            'ProductCreateView': ('create', 'new/'),
            'ProductUpdateView': ('update', 'change/<str:pk>/'),
            'ProductDeleteView': 'delete/<str:pk>/',
            'ProductDetailView': '<str:pk>/',
        }

        urlpatterns = trims.paths_dict(views, trim_patterns)

        trim_patterns = dict(
            ListView='',
            CreateView=('create', 'new/'),
            UpdateView=('update', 'change/<str:pk>/'),
            DeleteView='delete/<str:pk>/',
            DetailView='<str:pk>/',
        )

        urlpatterns = trims.paths_dict(views, trim_patterns, view_prefix='Product')
        urlpatterns = trims.paths_dict(views, trim_patterns, view_prefix='People')

    """
    flag_class = View
    # r = ()
    # d = {}
    d = ()
    patterns = patterns or {}
    view_prefix = clean_str(view_prefix)
    url_pattern_prefix = clean_str(url_pattern_prefix)
    url_name_prefix = clean_str(url_name_prefix)

    patterns.update(kw)
    # print(f' Building {len(patterns)} patterns for {views}')
    for part_name, solution in patterns.items():
        class_name = f"{view_prefix}{part_name}"

        try:
            view = getattr(views, class_name)
        except AttributeError as e:

            if ignore_missing_views is False:
                raise e
            continue

        if (
            isinstance(
                solution,
                (
                    list,
                    tuple,
                ),
            )
            is False
        ):
            current_name = class_name.lower()
            last_name = inspect.getmro(view)[1].__name__
            app_name = MAPPED_CLASS.get(last_name, current_name)
            solution = (
                app_name,
                solution,
            )

        path_name, url, *extra = solution
        _urls = url
        if (
            isinstance(
                url,
                (
                    tuple,
                    list,
                ),
            )
            is False
        ):
            _urls = (url,)

        # print('  > ', f'{path_name: <30}', _urls)
        for _url in _urls:
            """Unpack 1 or more URLS to produce a unique function name
            for each url.
            """
            furl = f"{url_pattern_prefix}{_url}"
            fname = f"{url_name_prefix}{path_name}"
            if (fname in d) or (safe_prefix is True):
                fname = f"{part_name.lower()}-{fname}"
            entry = (
                furl,
                view,
            ) + tuple(extra)
            # d[fname] = entry
            d += ((fname, entry),)

    r = paths(d)

    return r


def template_view(url_string, html_path, name="template_view"):

    # urlpatterns += [
    #     trims.template_view('mockup/', 'mockup/crystal-geometries.html')
    # ]

    return path(url_string, TemplateView.as_view(template_name=html_path), name=name)


def as_templates(**props):
    """

    urlpatterns += trims.as_templates(
        geoms=('mockup/', 'mockup/crystal-geometries.html')
    )

    """
    return [template_view(*v, name=k) for k, v in props.items()]


def paths(path_dict):
    """
    urlpatterns = trims.paths(
        list=('', views.ProductListView,),
        create=('new/', views.ProductCreateView,),
        update=('change/<str:pk>/', views.ProductUpdateView,),
        delete=('delete/<str:pk>/', views.ProductDeleteView,),
        detail=('<str:pk>/', views.ProductDetailView,),
    )

    urlpatterns = [
        path('', views.ProductListView.as_view(), name='list'),
        path('new/', views.ProductCreateView.as_view(), name='create'),
        path('change/<str:pk>/', views.ProductUpdateView.as_view(), name='update'),
        path('delete/<str:pk>/', views.ProductDeleteView.as_view(), name='delete'),
        path('<str:pk>/', views.ProductDetailView.as_view(), name='detail'),
    ]

    """
    flag_class = View
    r = ()
    for name, params in path_dict:  # dict(path_dict).items():
        (url, unit, *extra) = params
        func = unit
        if isinstance(func, tuple) is False:
            if inspect.isfunction(func) is False:

                try:
                    mros = inspect.getmro(unit)
                except Exception as exc:
                    import pdb

                    pdb.set_trace()  # breakpoint 2d6a262ax //
                if flag_class in mros:
                    view_params = {k: v for d in extra for k, v in d.items()}
                    func = unit.as_view(**view_params)
        # print('Paths Making', url,name)
        p = path(url, func, name=name)
        r += (p,)

    return list(r)


def error_handlers(name, setup=None, template_dir=None):
    """Implement the urls for the given module name

    error_handers(__name__)
    error_handers(__name__, {
            400: 'trims.views.errors.handler400',
            403: 'trims.views.errors.handler403',
            404: 'trims.views.errors.handler404',
            500: 'trims.views.errors.handler500',
        }, template_dir='trim/errors/'
    )

    """

    defaults = {
        400: "trim.views.errors.handler400",
        403: "trim.views.errors.handler403",
        404: "trim.views.errors.handler404",
        500: "trim.views.errors.handler500",
    }
    default_template_dir = "trim/errors/"
    template_dir = template_dir or default_template_dir
    defaults.update(setup or {})

    # rebuild the handler function names and set into the target module.
    _package, _module = name.split(".")
    # class_module_name = f'{_package}.views'
    module = sys.modules[name]

    # Push the new handler name into the targt module (shoppinglist.urls)
    for error_num, handler_path in defaults.items():
        handler_name = f"handler{error_num}"
        setattr(module, handler_name, handler_path)



import random
import string


def rand_str(length=6):
    """
    Generate a random string of uppercase letters and digits.

    Creates a random string composed of uppercase ASCII letters (A-Z) and
    digits (0-9). Useful for generating random identifiers, tokens, or
    temporary names.

    Args:
        length (int, optional): The length of the random string to generate.
            Defaults to 6.

    Returns:
        str: A random string of the specified length containing only uppercase
            letters and digits.

    Examples:
        Generate a 6-character random string (default):

            >>> token = rand_str()
            >>> len(token)
            6

        Generate a 12-character random string:

            >>> token = rand_str(12)
            >>> len(token)
            12

        Use for temporary file names:

            >>> filename = f"temp_{rand_str(8)}.txt"
            >>> # Example output: "temp_A7X9K2D1.txt"

    Notes:
        - The function uses `random.choices()` which allows for repeated characters
        - Character pool includes: A-Z (26 letters) and 0-9 (10 digits)
        - For cryptographically secure random strings, use `secrets.token_urlsafe()`
          instead
    """
    choices = random.choices(string.ascii_uppercase + string.digits, k=length)
    return "".join(choices)



import mimetypes
import os
import re
from wsgiref.util import FileWrapper

from django.http import StreamingHttpResponse
from django.utils.encoding import smart_str

range_re = re.compile(r"bytes\s*=\s*(\d+)\s*-\s*(\d*)", re.I)


def stream(request, path):
    """
    Stream a file to the client with support for HTTP range requests.

    This function handles file streaming with support for partial content requests
    (HTTP 206) using the Range header. It's useful for streaming large files,
    videos, or audio where clients may request specific byte ranges.

    Args:
        request (HttpRequest): The Django HTTP request object containing headers
            and metadata for the file request.
        path (str): The filesystem path to the file that should be streamed.

    Returns:
        StreamingHttpResponse: A Django streaming response object with appropriate
            headers set for content type, length, and range information.

    Raises:
        OSError: If the file at the specified path cannot be accessed or opened.
        FileNotFoundError: If the file at the specified path does not exist.

    Notes:
        - Supports HTTP Range requests (RFC 7233) for partial content delivery
        - Automatically detects content type using mimetypes
        - Falls back to 'application/octet-stream' if content type cannot be determined
        - Returns status 206 (Partial Content) for range requests
        - Returns status 200 for full file requests
        - Sets 'Accept-Ranges: bytes' to indicate range request support

    Examples:

        from django.views import View
        class MyFileView(View):
            def get(self, request, *args, **kwargs):
                file_path = '/path/to/my/largefile.mp4'
                return stream(request, file_path)

    Security Warning:
        This function does not perform any path validation or authorization checks.
        Ensure proper validation is implemented to prevent directory traversal
        attacks and unauthorized file access before calling this function.
    """
    range_header = request.META.get("HTTP_RANGE", "").strip()
    range_match = range_re.match(range_header)
    size = os.path.getsize(path)
    content_type, encoding = mimetypes.guess_type(path)
    content_type = content_type or "application/octet-stream"
    if range_match:
        first_byte, last_byte = range_match.groups()
        first_byte = int(first_byte) if first_byte else 0
        last_byte = int(last_byte) if last_byte else size - 1
        if last_byte >= size:
            last_byte = size - 1
        length = last_byte - first_byte + 1
        resp = StreamingHttpResponse(
            RangeFileWrapper(open(path, "rb"), offset=first_byte, length=length),
            status=206,
            content_type=content_type,
        )
        resp["Content-Length"] = str(length)
        resp["Content-Range"] = "bytes %s-%s/%s" % (first_byte, last_byte, size)
    else:
        resp = StreamingHttpResponse(
            FileWrapper(open(path, "rb")), content_type=content_type
        )
        resp["Content-Length"] = str(size)
    resp["Accept-Ranges"] = "bytes"
    return resp


def stream(request, path):
    """
    Stream a file to the client with support for HTTP range requests.

    This function handles file streaming with support for partial content requests
    (HTTP 206) using the Range header. It's useful for streaming large files,
    videos, or audio where clients may request specific byte ranges.

    Args:
        request (HttpRequest): The Django HTTP request object containing headers
            and metadata for the file request.
        path (str): The filesystem path to the file that should be streamed.

    Returns:
        StreamingHttpResponse: A Django streaming response object with appropriate
            headers set for content type, length, and range information.

    Raises:
        OSError: If the file at the specified path cannot be accessed or opened.
        FileNotFoundError: If the file at the specified path does not exist.

    Notes:
        - Supports HTTP Range requests (RFC 7233) for partial content delivery
        - Automatically detects content type using mimetypes
        - Falls back to 'application/octet-stream' if content type cannot be determined
        - Returns status 206 (Partial Content) for range requests
        - Returns status 200 for full file requests
        - Sets 'Accept-Ranges: bytes' to indicate range request support
        - Content-Length headers are intentionally omitted in this version

    Examples:

        from django.views import View
        class MyFileView(View):
            def get(self, request, *args, **kwargs):
                file_path = '/path/to/my/largefile.mp4'
                return stream(request, file_path)

    Security Warning:
        This function does not perform any path validation or authorization checks.
        Ensure proper validation is implemented to prevent directory traversal
        attacks and unauthorized file access before calling this function.
    """
    range_header = request.META.get("HTTP_RANGE", "").strip()
    range_match = range_re.match(range_header)
    size = os.path.getsize(path)
    content_type, encoding = mimetypes.guess_type(path)
    content_type = content_type or "application/octet-stream"
    if range_match:
        first_byte, last_byte = range_match.groups()
        first_byte = int(first_byte) if first_byte else 0
        last_byte = int(last_byte) if last_byte else size - 1
        if last_byte >= size:
            last_byte = size - 1
        length = last_byte - first_byte + 1
        resp = StreamingHttpResponse(
            RangeFileWrapper(open(path, "rb"), offset=first_byte, length=length),
            status=206,
            content_type=content_type,
        )
        # resp['Content-Length'] = str(length)
        resp["Content-Range"] = "bytes %s-%s/%s" % (first_byte, last_byte, size)
    else:
        resp = StreamingHttpResponse(
            FileWrapper(open(path, "rb")), content_type=content_type
        )
        # resp['Content-Length'] = str(size)
    resp["Accept-Ranges"] = "bytes"
    return resp


class RangeFileWrapper(object):
    def __init__(self, stream, chunk_size=8192, offset=0, length=None):
        self.stream = stream
        self.stream.seek(offset, os.SEEK_SET)
        self.remaining = length
        self.chunk_size = chunk_size

    def close(self):
        if hasattr(self.stream, "close"):
            self.stream.close()

    def __iter__(self):
        return self

    def next(self):
        if self.remaining is None:
            # If remaining is None, we're reading the entire file.
            data = self.stream.read(self.chunk_size)
            if data:
                return data
            raise StopIteration()

        if self.remaining <= 0:
            raise StopIteration()

        data = self.stream.read(min(self.remaining, self.chunk_size))
        if not data:
            raise StopIteration()

        self.remaining -= len(data)
        return data


def streamfile_response(
    real_filepath,
    output_filename,
    chunk_size=8192,
    content_type=None,
    range_header=None,
):
    """Generate a StreamingHttpResponse for the given `real_filepath`. Return
    directly to the client as the response object, e.g: from `get()`.

        response = streamfile_response('/real/path.zip', 'output_name.zip')

    Elements applied:

    + A wsgi FileWrapper
    + Auto content_type
    + StreamingHttpResponse response object
    + Content-Length
    + Content-Disposition
    + X-Sendfile
    """
    filename = os.path.basename(real_filepath)
    size = os.path.getsize(real_filepath)
    handle = open(real_filepath, "rb")
    # if ranges was given, seek the handle up to the point.
    wrapper = FileWrapper(handle, chunk_size)
    content_type = content_type or mimetypes.guess_type(real_filepath)[0]
    safe_filename = smart_str(output_filename)

    # You can also set any other required headers: Cache-Control, etc.
    response_pv = {
        # for bytes ranging partial downloads
        "Accept-Ranges": "bytes",
        "Content-Disposition": f"attachment; filename={safe_filename}",
        "X-Sendfile": smart_str(real_filepath),
    }

    status_ref = {}
    # range_header = request.META.get('HTTP_RANGE', '').strip()
    range_match = range_re.match(range_header or "")
    if range_match:
        status_ref = {
            "status": 206,
        }

        first_byte, last_byte = range_match.groups()
        first_byte = int(first_byte) if first_byte else 0
        last_byte = int(last_byte) if last_byte else size - 1
        if last_byte >= size:
            last_byte = size - 1
        length = last_byte - first_byte + 1
        response_pv["Content-Range"] = "bytes %s-%s/%s" % (first_byte, last_byte, size)

        size = length
        wrapper = RangeFileWrapper(handle, offset=first_byte, length=length)

    # response = HttpResponse(mimetype='application/force-download')
    response = StreamingHttpResponse(
        wrapper,
        **status_ref,
        content_type=content_type,
        # mimetype='application/force-download',
    )

    response["Content-Length"] = str(size)

    for k, v in response_pv.items():
        response[k] = v
    return response



from django import template
from loguru import logger

from .base import SlotList, parse_until

log = logger.debug


def slot_into_parent(parser, slotnode, slotlist_name, parent_node_name="wrap"):
    for node_name, token_node in parser.command_stack[::-1]:
        # push this into the wrap above.
        if node_name == parent_node_name:
            log(f'Pushing "{slotnode}" into parent "{node_name}".{slotlist_name}')
            if slotlist_name == "slotdefine":
                import pdb

                pdb.set_trace()  # breakpoint 5ae4487e //

            getattr(token_node, slotlist_name).add(slotnode)


# @register.tag(name="slot.define")
def do_define_slot(parser, token):
    """A definition slot to apply within the wrap target

        {% slot.define "name" %}content will be replaced{% endslot %}

    The wrap targets this slot

        {% wrap 'foo.html' %}
            {% slot "name" %}
                Content to replace the slot.
            {% endslot %}
        {% endwrap %}
    """
    nodelist, splits, extra = parse_until(parser, token, ("endslot",))
    res = DefineSlotNode(nodelist, splits, **extra)
    slot_into_parent(parser, res, "slotdefine")
    return res

    # for node_name, token_node in parser.command_stack[::-1]:
    #     # push this into the wrap above.
    #     if node_name == 'wrap':
    #         log('  _dis_ "slot.define" discovered parent (wrap)', token_node)
    #         token_node.slotdefine.add(res,)

    # return res


# @register.tag(name="slot")
def do_slot(parser, token):
    """A `{% slot %} {% endslot %}`  token to capture a placeholder replacement
    for a wrap. The HTML within is applied to the target slot.

    Parse the content and return a SlotNode. This new node is applied to the
    first discovered `{% wrap %}` parent.

    return the new SlotNode
    ---

    The appliance captures the nodes within within the block, then
    iterates backward through the parser node stack until the first
    `wrap` node.
    This new SlotNode instance is pushed into the wrap token `slotinfo`
    list, later inspected when the Wrap is rendered.

    Note: the slotinfo is within the _token_ `token_node` and not the _Node_,
    as the target _wrap_ node may not exist yet (being a currently unrendered
    token).
    """
    nodelist, tokens, extra = parse_until(parser, token, ("endslot",))
    res = SlotNode(nodelist, tokens, **extra)
    slot_into_parent(parser, res, "slotinfo")
    return res


from django.template.base import VariableDoesNotExist


class DefineSlotNode(template.Node):
    """The slot definition, within the main wrap.html."""

    def __init__(self, nodelist, tokens, *a, **kw):
        self.nodelist = nodelist
        first_token = tokens[0] if len(tokens) > 0 else self.get_default_name()
        self.first_token = first_token
        self.token_slot_name = template.Variable(first_token)
        self.tokens = tokens
        self.extra_context = kw
        self.args = a
        log(
            f" _cre_ DefineSlotNode({self.first_token}), {self.tokens} {self.args} {self.extra_context}"
        )

    def get_default_name(self):
        return "default"

    def get_slot_names(self, context=None):
        """
        context required _variables_ are to be processed
        """
        v = self.token_slot_name
        if context:
            try:
                v = v.resolve(context) if hasattr(v, "resolve") else v
            except VariableDoesNotExist:
                # the given string is not a variable. It could be a
                # const type (e.g. default) or a bad var.
                # Quietly fail
                return self.first_token
        return v

    def apply_parent(self, wrap_node):
        """set the parent wrapper."""
        self.extra_context["parent_wrap"] = wrap_node

    def resolve_extra_context(self, context):
        values = {}
        for key, val in self.extra_context.items():
            values[key] = val.resolve(context) if hasattr(val, "resolve") else val
        return values

    def get_slots(self, token, context):
        ## Grab 'name' from the wrap slots.
        parent_slots = self.extra_context["parent_wrap"].token.slotinfo
        matches = parent_slots.get_nodes(token, context)
        return matches

    def render(self, context):
        ## The content should be the content from the rendered slot info
        #
        # template_name = self.token_template_name.resolve(context)
        # t = context.template.engine.get_template(template_name)
        log(f"   _ren_ DefineSlotNode: {self.first_token}")

        # content applied within the definition,
        # before the parent slot implements any changes.
        default_slot_content = self.nodelist.render(context)
        content = default_slot_content

        matches = self.get_slots(self.first_token, context)
        log(
            f' . _ren_ Found: {len(matches)} matches during slot search for "{self.first_token}"'
        )

        if len(matches) > 0:
            ## The parent has a given a slot matching this definition name.
            # The content from the slot should be stripped from the wrap render
            # and applied here; replacing the `default_slot_content`
            ks = tuple(x[0] for x in matches)
            log(f"Found matches {ks}")
            # content = f'Found matches {ks}'
            vv = self.nodelist.render(context)
            for name, node in matches:
                # pump out first detection
                return node.nodelist.render(context)
        else:
            log("No matches")
        # we extend the original context; allowing this unit to be overwitten
        # if required.
        sub_ctx = {
            # Given to the definition slot
            "slot": {"rendered_content": content}
        }

        values = self.resolve_extra_context(context)
        with context.push(**sub_ctx, **values):
            # Context({'source': content}, autoescape=context.autoescape)
            return self.nodelist.render(context)


class SlotNode(template.Node):
    """A {% slot %}{% endslot %} assigns the given content to an allocated
    location with the parent {% wrap %}. The wrap maintains `slot.define` areas
    with a matching name.

            {% wrap "foo.html "%}
                {% slot "target" %}
                    <p>Content</p>
                {% endslot %}
            {% endwrap %}

    Functionally this doesn't render its content - as that task is managed by
    the definition slot.
    """

    def __init__(self, nodelist, tokens, *a, **kw):
        self.nodelist = nodelist
        first_token = tokens[0] if len(tokens) > 0 else self.get_default_name()
        self.first_token = first_token
        self.token_slot_name = template.Variable(first_token)
        self.tokens = tokens
        self.extra_context = kw
        self.args = a
        log(f" _cre_ SlotNode, {self.tokens} {self.args} {self.extra_context}")

    def get_default_name(self):
        return "default"

    def get_slot_names(self, context=None):
        """
        context required _variables_ are to be processed
        """
        v = self.token_slot_name
        if context:
            try:
                v = v.resolve(context) if hasattr(v, "resolve") else v
            except VariableDoesNotExist:
                # the given string is not a variable. It could be a
                # const type (e.g. default) or a bad var.
                # Quietly fail
                return self.first_token
        return v

    def render(self, context):
        # template_name = self.token_template_name.resolve(context)
        # t = context.template.engine.get_template(template_name)
        names = self.get_slot_names(context)
        # self.default_render(self, context)
        log(f"   _ren_ SlotNode, {names}")
        return f'<div class="color-gray">cut slot {names}</div>'

    def default_render(self, context):
        content = self.nodelist.render(context)
        # we extend the original context; allowing this unit to e overbwitten
        # if required.
        wrap = {
            "content": content,
        }
        #  Dynamic key for the root context, static key for the 'wrap' object.
        content_key = "content"
        sub_ctx = {}  # = {'wrap':wrap, content_key:content}
        values = {key: val.resolve(context) for key, val in self.extra_context.items()}
        with context.push(**sub_ctx, **values):
            # Context({'source': content}, autoescape=context.autoescape)
            return self.nodelist.render(context)



def cook(func):
    """Given a function, bake the functionality into trims,
    allowing for generic import across the app.

    Pre baked items are cooked immediately.
    """
    print("Cook", func)
